# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
# Imports
import asyncio
import time

import concurrent.futures
from concurrent.futures import Future

from dataclasses import dataclass
from typing import Union

import pytest

from tango import (
    ApiUtil,
    AttrReadEvent,
    AttrWriteType,
    AttrWrittenEvent,
    CmdDoneEvent,
    DevFailed,
    GreenMode,
    cb_sub_model,
    get_device_proxy,
)
from tango.server import Device, attribute, command
from tango.test_utils import DeviceTestContext, assert_close

A_BIT = 0.1


class ServerTest(Device):
    _value1 = 100
    _value2 = 200

    @attribute(access=AttrWriteType.READ_WRITE)
    def attr1(self) -> int:
        return self._value1

    @attr1.write
    def set_attr1(self, val):
        self._value1 = val

    @attribute(access=AttrWriteType.READ_WRITE)
    def attr2(self) -> int:
        return self._value2

    @attr2.write
    def set_attr2(self, val):
        self._value2 = val

    @attribute(dtype=int, access=AttrWriteType.READ)
    def attr_no_value(self):
        return

    @attribute(dtype=int, access=AttrWriteType.READ)
    def attr_exception(self):
        raise RuntimeError("Force exception for test")


@pytest.mark.asyncio
async def test_green_mode_kwarg_for_proxy_methods():
    with DeviceTestContext(ServerTest, device_name="test/test_device/1"):
        dev = await get_device_proxy("test/test_device/1", green_mode=GreenMode.Asyncio)

        # standard behavior for single attribute
        value = 12
        async_single_write_id = await dev.write_attribute_asynch("attr1", value)
        time.sleep(A_BIT)
        await dev.write_attribute_reply(async_single_write_id)

        async_single_read_id = await dev.read_attribute_asynch("attr1")
        time.sleep(A_BIT)
        attr = await dev.read_attribute_reply(async_single_read_id)
        assert value == attr.value

        # standard behavior for multiple attributes
        value = 34
        async_multiple_write_id = await dev.write_attributes_asynch([("attr1", value)])
        time.sleep(A_BIT)
        await dev.write_attributes_reply(async_multiple_write_id)

        async_multiple_read_id = await dev.read_attributes_asynch(["attr1"])
        time.sleep(A_BIT)
        attr = await dev.read_attributes_reply(async_multiple_read_id)
        assert value == attr[0].value

        # force sync execution for single attribute
        value = 56
        sync_single_write_id = dev.write_attribute_asynch(
            "attr1", value, green_mode=GreenMode.Synchronous
        )
        assert not isinstance(sync_single_write_id, asyncio.Future)
        time.sleep(A_BIT)
        dev.write_attribute_reply(
            sync_single_write_id, green_mode=GreenMode.Synchronous
        )

        sync_single_read_id = dev.read_attribute_asynch(
            "attr1", green_mode=GreenMode.Synchronous
        )
        assert not isinstance(sync_single_read_id, asyncio.Future)
        time.sleep(A_BIT)
        attr = dev.read_attribute_reply(
            sync_single_read_id, green_mode=GreenMode.Synchronous
        )
        assert value == attr.value

        # force sync execution for multiple attributes
        value = 78
        sync_multiple_write_id = dev.write_attributes_asynch(
            [("attr1", value)], green_mode=GreenMode.Synchronous
        )
        assert not isinstance(sync_multiple_write_id, asyncio.Future)
        time.sleep(A_BIT)
        dev.write_attributes_reply(
            sync_multiple_write_id, green_mode=GreenMode.Synchronous
        )

        sync_multiple_read_id = dev.read_attributes_asynch(
            ["attr1"], green_mode=GreenMode.Synchronous
        )
        assert not isinstance(sync_multiple_read_id, asyncio.Future)
        time.sleep(A_BIT)
        attr = dev.read_attributes_reply(
            sync_multiple_read_id, green_mode=GreenMode.Synchronous
        )
        assert value == attr[0].value


def test_async_write_with_attr_info():
    with DeviceTestContext(ServerTest) as proxy:

        attr_info = proxy.get_attribute_config_ex(["attr1", "attr2"])
        # reading single attribute with attr_info (no additional IO to fetch attr info during write method)
        value = 12
        async_single_write_id = proxy.write_attribute_asynch(attr_info[0], value)
        time.sleep(A_BIT)
        proxy.write_attribute_reply(async_single_write_id)

        async_single_read_id = proxy.read_attribute_asynch("attr1")
        time.sleep(A_BIT)
        attr = proxy.read_attribute_reply(async_single_read_id)
        assert value == attr.value

        # standard behavior for multiple attributes
        values = [34, 56]
        async_multiple_write_id = proxy.write_attributes_asynch(
            [(attr_info[0], values[0]), [attr_info[1], values[1]]]
        )
        time.sleep(A_BIT)
        proxy.write_attributes_reply(async_multiple_write_id)

        async_multiple_read_id = proxy.read_attributes_asynch(["attr1", "attr2"])
        time.sleep(A_BIT)
        attr = proxy.read_attributes_reply(async_multiple_read_id)
        assert values[0] == attr[0].value
        assert values[1] == attr[1].value


def test_async_attribute_polled():
    with DeviceTestContext(ServerTest) as proxy:
        # asynchronous write/read of single attribute
        single_write_id = proxy.write_attribute_asynch("attr1", 123)
        time.sleep(A_BIT)
        proxy.write_attribute_reply(single_write_id)

        single_read_id = proxy.read_attribute_asynch("attr1")
        time.sleep(A_BIT)
        attr = proxy.read_attribute_reply(single_read_id)
        assert 123 == attr.value

        # asynchronous write/read of multiple attributes
        multiple_write_id = proxy.write_attributes_asynch(
            [("attr1", 456), ("attr2", 789)]
        )
        time.sleep(A_BIT)
        proxy.write_attributes_reply(multiple_write_id)

        multiple_read_id = proxy.read_attributes_asynch(["attr1", "attr2"])
        time.sleep(A_BIT)
        attrs = proxy.read_attributes_reply(multiple_read_id)
        assert [456, 789] == [attr.value for attr in attrs]


def test_async_attribute_polled_no_return_value_or_exception():
    with DeviceTestContext(ServerTest) as proxy:
        read_id = proxy.read_attribute_asynch("attr_no_value")
        with pytest.raises(DevFailed, match="API_AttrValueNotSet"):
            proxy.read_attribute_reply(read_id, poll_timeout=int(A_BIT * 1000))

        read_id = proxy.read_attribute_asynch("attr_exception")
        with pytest.raises(DevFailed, match="RuntimeError"):
            proxy.read_attribute_reply(read_id, poll_timeout=int(A_BIT * 1000))

        multiple_read_id = proxy.read_attributes_asynch(
            ["attr1", "attr_no_value", "attr_exception"]
        )
        attr1, attr_no_value, attr_exception = proxy.read_attributes_reply(
            multiple_read_id, poll_timeout=int(A_BIT * 1000)
        )
        assert attr1.value == 100
        assert not attr1.has_failed
        assert attr_no_value.has_failed
        assert attr_exception.has_failed


@pytest.mark.parametrize("model", ["poll", "push"])
def test_async_attribute_with_callback(model):
    callbacks = []

    def write_callback(attr_written_event: AttrWrittenEvent):
        assert_close(attr_written_event.attr_names, ["attr1", "attr2"])
        assert attr_written_event.device == proxy
        assert not attr_written_event.err
        callbacks.append(attr_written_event)

    def read_callback(attr_read_event: AttrReadEvent):
        assert_close(attr_read_event.attr_names, ["attr1", "attr2"])
        assert_close([attr.value for attr in attr_read_event.argout], [123, 456])
        assert attr_read_event.device == proxy
        assert not attr_read_event.err
        callbacks.append(attr_read_event)

    api_util = ApiUtil.instance()
    api_util.set_asynch_cb_sub_model(
        cb_sub_model.PUSH_CALLBACK if model == "push" else cb_sub_model.PULL_CALLBACK
    )

    with DeviceTestContext(ServerTest) as proxy:
        # asynchronous write/read of multiple attributes
        proxy.write_attributes_asynch([("attr1", 123), ("attr2", 456)], write_callback)
        time.sleep(A_BIT)
        proxy.read_attributes_asynch(["attr1", "attr2"], read_callback)

        if model == "poll":
            api_util.get_asynch_replies(500)
        else:
            time.sleep(0.5)

        assert len(callbacks) == 2


def test_async_command_polled():
    class TestDevice(Device):
        @command(dtype_in=int, dtype_out=int)
        def identity(self, arg):
            return arg

    with DeviceTestContext(TestDevice) as proxy:
        eid = proxy.command_inout_asynch("identity", 123)
        assert 123 == proxy.command_inout_reply(eid, timeout=500)


class ServerForAsynchClients(Device):
    @command(dtype_in=int, dtype_out=int)
    def cmd_ok(self, arg):
        return arg

    @command(dtype_in=int, dtype_out=int)
    def cmd_timeout(self, arg):
        time.sleep(0.2)
        return arg

    @command(dtype_in=int, dtype_out=int)
    def cmd_exception(self, arg):
        raise Exception("Intentional exception")

    @attribute(dtype=int)
    def attr_ok(self):
        return 123

    @attr_ok.setter
    def attr_ok(self, value):
        pass

    @attribute(dtype=int)
    def attr_timeout(self):
        time.sleep(0.2)
        return 123

    @attr_timeout.setter
    def attr_timeout(self, value):
        time.sleep(0.2)
        pass

    @attribute(dtype=int)
    def attr_exception(self):
        raise Exception("Intentional exception")

    @attr_exception.setter
    def attr_exception(self, value):
        raise Exception("Intentional exception")


@pytest.mark.parametrize(
    "cmd,argin,argout,err,err_str",
    [
        ("cmd_ok", 123, 123, False, ""),
        ("cmd_timeout", 123, None, True, "TRANSIENT_CallTimedout"),
        ("cmd_exception", 123, None, True, "Intentional exception"),
    ],
)
def test_async_command_with_polled_callback(cmd, argin, argout, err, err_str):
    api_util = ApiUtil.instance()
    api_util.set_asynch_cb_sub_model(cb_sub_model.PULL_CALLBACK)

    with DeviceTestContext(ServerForAsynchClients, process=True) as proxy:
        future = Future()
        proxy.set_timeout_millis(
            150
        )  # this timeout does not have any influence on get_asynch_replies behaviour
        proxy.command_inout_asynch(cmd, argin, future.set_result)
        api_util.get_asynch_replies(500)  # this timeout is the one that matters
        result: CmdDoneEvent = future.result()
        assert result.err == err
        if not err:
            assert result.argout == argout
            assert len(result.errors) == 0
        else:
            with pytest.raises(DevFailed, match="API_EmptyDeviceData"):
                _ = result.argout
            assert err_str in str(result.errors[0])


@dataclass
class AttrReading:
    has_failed: bool
    value: Union[int, None]


@pytest.mark.parametrize(
    "attr,argout,err",
    [
        ("attr_ok", [AttrReading(has_failed=False, value=123)], False),
        ("attr_timeout", [], True),
        ("attr_exception", [AttrReading(has_failed=True, value=None)], True),
    ],
)
def test_async_attribute_read_with_polled_callback(attr, argout, err):
    api_util = ApiUtil.instance()
    api_util.set_asynch_cb_sub_model(cb_sub_model.PULL_CALLBACK)

    with DeviceTestContext(ServerForAsynchClients, process=True) as proxy:
        future = Future()
        proxy.set_timeout_millis(150)
        proxy.read_attribute_asynch(attr, future.set_result)
        api_util.get_asynch_replies(500)
        result: AttrReadEvent = future.result()
        assert result.err == err
        assert len(result.argout) == len(argout)
        # compare values of returned attributes
        for actual, expected in zip(result.argout, argout):
            assert actual.has_failed == expected.has_failed
            assert actual.value == expected.value


@pytest.mark.parametrize(
    "attr,err,num_errors,err_str",
    [
        ("attr_ok", False, 0, ""),
        ("attr_timeout", True, 0, ""),
        ("attr_exception", True, 1, "Intentional exception"),
    ],
)
def test_async_attribute_write_with_polled_callback(attr, err, num_errors, err_str):
    api_util = ApiUtil.instance()
    api_util.set_asynch_cb_sub_model(cb_sub_model.PULL_CALLBACK)

    with DeviceTestContext(ServerForAsynchClients, process=True) as proxy:
        future = Future()
        proxy.set_timeout_millis(150)
        proxy.write_attribute_asynch(attr, 123, future.set_result)
        # Future should be emtpy if we have not pulled replies
        with pytest.raises((concurrent.futures.TimeoutError, TimeoutError)):
            _ = future.result(timeout=0.1)

        api_util.get_asynch_replies(500)
        result: AttrWrittenEvent = future.result(timeout=0.1)

        assert result.err == err
        assert len(result.errors.err_list) == num_errors
        if num_errors > 0:
            assert err_str in str(result.errors.err_list[0].err_stack[0])


@pytest.mark.parametrize(
    "cmd,argin,argout,err,err_str",
    [
        ("cmd_ok", 123, 123, False, ""),
        ("cmd_timeout", 123, None, True, "TRANSIENT_CallTimedout"),
        ("cmd_exception", 123, None, True, "Intentional exception"),
    ],
)
def test_async_command_with_pushed_callback(cmd, argin, argout, err, err_str):
    api_util = ApiUtil.instance()
    api_util.set_asynch_cb_sub_model(cb_sub_model.PUSH_CALLBACK)

    with DeviceTestContext(ServerForAsynchClients, process=True) as proxy:
        future = Future()
        proxy.set_timeout_millis(100)
        proxy.command_inout_asynch(cmd, argin, future.set_result)
        result: CmdDoneEvent = future.result(timeout=5)
        assert result.err == err
        if not err:
            assert result.argout == argout
            assert len(result.errors) == 0
        else:
            with pytest.raises(DevFailed, match="API_EmptyDeviceData"):
                _ = result.argout
            assert err_str in str(result.errors[0])


@pytest.mark.parametrize(
    "attr,argout,err",
    [
        ("attr_ok", [AttrReading(has_failed=False, value=123)], False),
        ("attr_timeout", [], True),
        ("attr_exception", [AttrReading(has_failed=True, value=None)], True),
    ],
)
def test_async_attribute_read_with_pushed_callback(attr, argout, err):
    api_util = ApiUtil.instance()
    api_util.set_asynch_cb_sub_model(cb_sub_model.PUSH_CALLBACK)

    with DeviceTestContext(ServerForAsynchClients, process=True) as proxy:
        future = Future()
        proxy.set_timeout_millis(150)
        proxy.read_attribute_asynch(attr, future.set_result)
        result: AttrReadEvent = future.result(timeout=5)
        assert result.err == err
        assert len(result.argout) == len(argout)
        # compare values of returned attributes
        for actual, expected in zip(result.argout, argout):
            assert actual.has_failed == expected.has_failed
            assert actual.value == expected.value


@pytest.mark.parametrize(
    "attr,err,num_errors,err_str",
    [
        ("attr_ok", False, 0, ""),
        ("attr_timeout", True, 0, ""),
        ("attr_exception", True, 1, "Intentional exception"),
    ],
)
def test_async_attribute_write_with_pushed_callback(attr, err, num_errors, err_str):
    api_util = ApiUtil.instance()
    api_util.set_asynch_cb_sub_model(cb_sub_model.PUSH_CALLBACK)

    with DeviceTestContext(ServerForAsynchClients, process=True) as proxy:
        future = Future()
        proxy.set_timeout_millis(150)
        proxy.write_attribute_asynch(attr, 123, future.set_result)
        result: AttrWrittenEvent = future.result(timeout=5)
        assert result.err == err
        assert len(result.errors.err_list) == num_errors
        if num_errors > 0:
            assert err_str in str(result.errors.err_list[0].err_stack[0])


def test_async_exception_in_callback():
    api_util = ApiUtil.instance()
    api_util.set_asynch_cb_sub_model(cb_sub_model.PUSH_CALLBACK)

    def callback(_event):
        raise Exception("Some exception")

    with DeviceTestContext(ServerForAsynchClients) as proxy:
        proxy.read_attribute_asynch("attr_ok", callback)
        time.sleep(0.2)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

# Imports
import functools
import asyncio
import inspect
import collections
import os
import types
import concurrent.futures
import warnings
import threading

from asyncio import futures, coroutines
from asyncio.tasks import ensure_future
from typing import Callable

# Tango imports
from tango.green import AbstractExecutor
from tango.utils import (
    _get_current_otel_context,
    _get_non_tango_source_location,
    _is_coroutine_function,
    PyTangoThreadPoolExecutor,
)

__all__ = (
    "AsyncioExecutor",
    "get_global_executor",
    "set_global_executor",
    "_switch_global_executor_to_thread",
)

_ALREADY_WARNED_FUNCTIONS = []


# Function removed from Python 3.11
# Taken from https://github.com/python/cpython/blob/3.10/Lib/asyncio/coroutines.py
# (without the _DEBUG part)
# Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
# 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022 Python Software Foundation;
# All Rights Reserved
def _coroutine(func):
    """Decorator to mark coroutines.
    If the coroutine is not yielded from before it is destroyed,
    an error message is logged.
    """
    if inspect.iscoroutinefunction(func):
        return func

    if inspect.isgeneratorfunction(func):
        coro = func
    else:

        @functools.wraps(func)
        def coro(*args, **kw):
            res = func(*args, **kw)
            if asyncio.isfuture(res) or inspect.isgenerator(res):
                res = yield from res
            else:
                # If 'res' is an awaitable, run it.
                try:
                    await_meth = res.__await__
                except AttributeError:
                    pass
                else:
                    if isinstance(res, collections.abc.Awaitable):
                        res = yield from await_meth()
            return res

    coro = types.coroutine(coro)
    wrapper = coro
    wrapper._is_coroutine = (
        asyncio.coroutines._is_coroutine
    )  # For iscoroutinefunction().
    return wrapper


# In Python 3.12 the legacy generator-based coroutines are not allowed for execution anymore
# Here we use modified run_coroutine_threadsafe method, which still can execute them
# Taken from https://github.com/python/cpython/blob/3.12/Lib/asyncio/tasks.py
# Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
# 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022, 2023 Python Software Foundation;
# All Rights Reserved
def run_coroutine_threadsafe(coro, loop):
    """Submit a coroutine object to a given event loop.

    Return a concurrent.futures.Future to access the result.
    """
    if not coroutines.iscoroutine(coro) and not types.GeneratorType:
        raise TypeError("A coroutine object is required")

    future = concurrent.futures.Future()

    def callback():
        try:
            futures._chain_future(ensure_future(coro, loop=loop), future)
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as exc:
            if future.set_running_or_notify_cancel():
                future.set_exception(exc)
            raise

    loop.call_soon_threadsafe(callback)
    return future


_PYTANGOTHREADPOOLEXECUTOR = None


def get_thread_pool_executor():
    global _PYTANGOTHREADPOOLEXECUTOR
    if _PYTANGOTHREADPOOLEXECUTOR is None:
        _PYTANGOTHREADPOOLEXECUTOR = PyTangoThreadPoolExecutor(
            thread_name_prefix="_PyTangoThreadPoolExecutor"
        )
    return _PYTANGOTHREADPOOLEXECUTOR


# Global executor
_MAIN_EXECUTOR = None
_THREAD_EXECUTORS = {}


def _switch_global_executor_to_thread():
    """
    internal PyTango function, use only if you sure, what you are doing!
    Used for correct behavior of TestDeviceContext
    Checks, that global executor belongs to the caller thread, and,
    if not - creates a new one and saves it as a new global
    """
    global _MAIN_EXECUTOR
    if _MAIN_EXECUTOR is not None and not _MAIN_EXECUTOR.in_executor_context():
        # we save current executor in the known subthread executors to be used later
        _THREAD_EXECUTORS[_MAIN_EXECUTOR.get_ident()] = _MAIN_EXECUTOR
        _MAIN_EXECUTOR = AsyncioExecutor()


def get_global_executor():
    global _MAIN_EXECUTOR
    if _MAIN_EXECUTOR is None:
        _MAIN_EXECUTOR = AsyncioExecutor()
    # the following patch is used for correct behavior of TestDeviceContext,
    # which has two different executors for main and device threads
    if not _MAIN_EXECUTOR.in_executor_context():
        ident = threading.get_ident(), os.getpid()
        if ident in _THREAD_EXECUTORS:
            return _THREAD_EXECUTORS[ident]

    return _MAIN_EXECUTOR


def set_global_executor(executor):
    global _MAIN_EXECUTOR
    _MAIN_EXECUTOR = executor


def _get_function_name(fn: Callable) -> str:
    if hasattr(fn, "__qualname__"):
        return fn.__qualname__
    elif hasattr(fn, "__name__"):
        return fn.__name__

    return f"{fn}"


# Asyncio executor
class AsyncioExecutor(AbstractExecutor):
    """Asyncio tango executor"""

    asynchronous = True
    default_wait = False

    def __init__(self, loop=None, subexecutor=None):
        super().__init__()
        if loop is None:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
        self.loop = loop
        self.subexecutor = (
            subexecutor if subexecutor is not None else get_thread_pool_executor()
        )

    def delegate(self, fn, *args, **kwargs):
        """Return the given operation as an asyncio future."""
        if hasattr(fn, "__trace_kwargs__"):
            kwargs["trace_location"] = _get_non_tango_source_location()
            kwargs["trace_context"] = _get_current_otel_context()
        callback = functools.partial(fn, *args, **kwargs)
        coro = self.loop.run_in_executor(self.subexecutor, callback)
        return asyncio.ensure_future(coro)

    def access(self, accessor, timeout=None):
        """Return a result from an asyncio future."""
        if self.loop.is_running():
            raise RuntimeError("Loop is already running")
        coro = asyncio.wait_for(accessor, timeout)
        return self.loop.run_until_complete(coro)

    def submit(self, fn, *args, **kwargs):
        """Submit an operation"""
        if _is_coroutine_function(fn):
            return run_coroutine_threadsafe(fn(*args, **kwargs), self.loop)
        else:
            # we leave this part of the code to support legacy servers
            name = _get_function_name(fn)
            if name not in _ALREADY_WARNED_FUNCTIONS:
                _ALREADY_WARNED_FUNCTIONS.append(name)
                warnings.warn(
                    f"Sync {name} function called: support of "
                    f"sync functions in PyTango's Asyncio mode is "
                    f"deprecated. Use 'async def' instead of 'def'.",
                    DeprecationWarning,
                )
            corofn = _coroutine(lambda: fn(*args, **kwargs))
            return run_coroutine_threadsafe(corofn(), self.loop)

    def execute(self, fn, *args, **kwargs):
        """Execute an operation and return the result."""
        if self.in_executor_context():
            if _is_coroutine_function(fn):
                return fn(*args, **kwargs)
            else:
                # we leave this part of the code to support legacy servers
                name = _get_function_name(fn)
                if name not in _ALREADY_WARNED_FUNCTIONS:
                    _ALREADY_WARNED_FUNCTIONS.append(name)
                    warnings.warn(
                        f"Sync {name} function called: support of "
                        f"sync functions in PyTango's Asyncio mode is "
                        f"deprecated. Use 'async def' instead of 'def'.",
                        DeprecationWarning,
                    )
                corofn = _coroutine(lambda: fn(*args, **kwargs))
                return corofn()
        future = self.submit(fn, *args, **kwargs)
        return future.result()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("device_attribute_init",)

__docformat__ = "restructuredtext"

import copy

from tango import DeviceAttribute, ExtractAs


def __DeviceAttribute__get_data(self):
    return self.get_data_raw().extract()


def __DeviceAttribute__init(self, da=None):
    if da is None:
        DeviceAttribute.__init_orig(self)
    else:
        DeviceAttribute.__init_orig(self, da)
        try:
            self.value = copy.deepcopy(da.value)
        except Exception:
            pass
        try:
            self.w_value = copy.deepcopy(da.w_value)
        except Exception:
            pass
        try:
            self.scalar_w_value = da.scalar_w_value
        except Exception:
            pass
        self.type = da.type
        self.is_empty = da.is_empty
        self.has_failed = da.has_failed


def device_attribute_init():
    DeviceAttribute.__init_orig = DeviceAttribute.__init__
    DeviceAttribute.__init__ = __DeviceAttribute__init
    DeviceAttribute.ExtractAs = ExtractAs



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# client example: reading an attribute, also with simplified syntax

import sys
from tango import DeviceProxy, DevFailed, Except

dev_name = "sys/tg_test/1"
attr_name = "double_scalar"

# step 1 - create the device proxy
try:
    dev = DeviceProxy(dev_name)
    print(f"proxy for {dev_name} created")
except DevFailed as ex:
    Except.print_exception(ex)
    sys.exit(1)

# step 2 - read the attribute
try:
    da = dev.read_attribute(attr_name)
    print(f"attribute.name: {da.name}")
    print(f"attribute.timestamp: {da.get_date()}")
    print(f"attribute.quality: {da.quality}")
    print(f"attribute.dim_x: {da.dim_x}")
    print(f"attribute.type: {da.type}")
    print(f"attribute.value: {da.value}")
except DevFailed as ex:
    Except.print_exception(ex)
    print("failed to read attribute")
    sys.exit(1)  # simplistic error handling

# set a value with Jive and check the result here
try:
    # read the attribute again
    da = dev.read_attribute(attr_name)
    print(f"attribute.value    (reading): {da.value}")
    print(f"attribute.w_value  (setting): {da.w_value}")
except DevFailed as ex:
    Except.print_exception(ex)
    print("failed to read attribute")
    sys.exit(1)

"""
Typical output:

➜  training git:(develop) ✗ docker-compose exec cli /training/client/reading02.py
proxy for sys/tg_test/1 created
attribute.name: double_scalar
attribute.timestamp: 2021-06-30 20:07:50.801692
attribute.quality: ATTR_VALID
attribute.dim_x: 1
attribute.type: DevDouble
attribute.value: -226.03415499895206
attribute.value    (reading): -226.03415499895206
attribute.w_value  (setting): 0.0
"""



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import contextlib
import functools
import logging
import os
import re
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from itertools import groupby
from operator import itemgetter

import tango
from tango.databaseds.db_errors import (
    DB_DeviceNotDefined,
    DB_IncorrectArguments,
    DB_IncorrectServerName,
    DB_SQLError,
)
from tango.test_context import parse_ior

th_exc = tango.Except.throw_exception
Executor = ThreadPoolExecutor(1)
cursor_lock = threading.RLock()


def log_me(f):
    @functools.wraps(f)
    def inner(self, *args, **kwargs):
        logging.info(f"Entering '{f.__name__}', args={args}, kwargs={kwargs}")
        try:
            t0 = time.time()
            result = f(self, *args, **kwargs)
            dt = time.time() - t0
            logging.info(
                f"Leaving '{f.__name__}' successfully, took {dt:.3f} s, returned({result})"
            )
            return result
        except tango.DevFailed:
            # This should be a "normal" error, intended for the client
            # No need to log as an error.
            logging.debug(f"Tango exception raised in {f.__name__}")
            raise
        except Exception:
            # This probably indicates a bug?
            logging.exception(f"Error in {f.__name__}")
            raise

    return inner


def get_create_db_statements():
    this_dir = os.path.dirname(__file__)
    statements = []
    with open(os.path.join(this_dir, "create_db_tables.sql")) as f:
        lines = f.readlines()
    # strip comments
    lines = (line for line in lines if not line.startswith("#"))
    lines = (line for line in lines if not line.lower().strip().startswith("key"))
    lines = (line for line in lines if not line.lower().strip().startswith("key"))
    lines = "".join(lines)
    lines = lines.replace("ENGINE=MyISAM", "")
    statements += lines.split(";")

    with open(os.path.join(this_dir, "create_db.sql")) as f:
        lines = f.readlines()
    # strip comments
    lines = (line for line in lines if not line.lower().startswith("#"))
    lines = (line for line in lines if not line.lower().startswith("create database"))
    lines = (line for line in lines if not line.lower().startswith("use"))
    lines = (line for line in lines if not line.lower().startswith("source"))
    lines = "".join(lines)
    statements += lines.split(";")

    return statements


def replace_wildcard(text):
    # escape '%' with '\'
    text = text.replace("%", "\\%")
    # escape '_' with '\'
    text = text.replace("_", "\\_")
    # escape '"' with '\'
    text = text.replace('"', '\\"')
    # escape ''' with '\'
    text = text.replace("'", "\\'")
    # replace '*' with '%'
    text = text.replace("*", "%")
    return text


def use_cursor(f):
    @functools.wraps(f)
    def wrap(self, *args, **kwargs):
        with cursor_lock:
            has_cursor = "cursor" in kwargs
            cursor = kwargs.pop("cursor", None)
            if not has_cursor:
                cursor = Executor.submit(self._get_cursor).result()
            self.cursor = cursor
            try:
                ret = Executor.submit(f, self, *args, **kwargs).result()
                if not has_cursor:
                    Executor.submit(self.cursor.connection.commit).result()
                return ret
            finally:
                if not has_cursor:
                    Executor.submit(self.cursor.close).result()
                    with contextlib.suppress(AttributeError):
                        del self.cursor

    return wrap


def regexp(expr, item):
    """Implement REGEXP function for MySQL compatibility"""
    # TODO may not be 100% compatible, as there are different regexp "flavors"
    if not isinstance(item, str):
        # Not sure when this can happen, but at least NULL (None) does
        return False
    return re.search(expr, item) is not None


class SqlDatabase:
    DB_API_NAME = "sqlite3"  # Default implementation

    def __init__(
        self,
        name,
        db_name=None,
        history_depth=10,
        fire_to_starter=False,  # TODO seems broken on pytango < 10
    ):
        self._db_api = None
        self._db_conn = None
        self.name = name
        self.dev_name = "sys/database/" + name
        if db_name is None:
            self.db_name = os.environ.get("PYTANGO_DATABASE_NAME", "tango_database.db")
        else:
            self.db_name = db_name
        self.history_depth = history_depth
        self.fire_to_starter = fire_to_starter
        self._logger = logging.getLogger(self.__class__.__name__)
        self._debug = self._logger.debug
        self._info = self._logger.info
        self._warn = self._logger.warn
        self._error = self._logger.error
        self._critical = self._logger.critical
        self._initialize()

    def close_db(self):
        if self._db_conn is not None:
            self._db_conn.commit()
            self._db_conn.close()
        self._db_api = None
        self._db_conn = None

    def get_db_api(self):
        if self._db_api is None:
            self._db_api = __import__(self.DB_API_NAME)
        return self._db_api

    @property
    def db_api(self):
        return self.get_db_api()

    @property
    def db_conn(self):
        if self._db_conn is None:
            self._db_conn = self.db_api.connect(self.db_name)
            if logging.root.level == logging.DEBUG:
                self._db_conn.set_trace_callback(lambda q: print(f"  Query: {q}"))
            self._db_conn.row_factory = self.db_api.Row
            # For MySQL compatibility, add REGEXP function
            self._db_conn.create_function("REGEXP", 2, regexp)
        return self._db_conn

    def _get_cursor(self):
        return self.db_conn.cursor()

    def _initialize(self):
        self._info(
            "Initializing database %r (%s)...",
            self.db_name,
            os.path.isfile(self.db_name),
        )
        if not os.path.isfile(self.db_name):
            self._create_db()
        else:
            # trigger connection
            self._trigger_connection()

    @use_cursor
    def _trigger_connection(self):
        return self.db_conn

    @use_cursor
    def _create_db(self):
        self._info("Creating database...")
        statements = get_create_db_statements()
        cursor = self.cursor
        for statement in statements:
            cursor.execute(statement)

    def _get_id(self, name, cursor):
        name += "_history_id"
        _id = cursor.execute(f"SELECT id FROM {name}").fetchone()[0] + 1
        cursor.execute(f"UPDATE {name} SET id={_id}")
        return _id

    def _purge_att_property(self, table, field, obj, attr, name, cursor):
        cursor.execute(
            f"SELECT DISTINCT id FROM {table} WHERE {field} = ? AND name = ? AND "
            + "attribute = ? ORDER BY date",
            (obj, name, attr),
        )
        rows = cursor.fetchall()
        to_del = len(rows) - self.history_depth
        if to_del > 0:
            for row in rows[:to_del]:
                cursor.execute(f"DELETE FROM {table} WHERE id=?", (row[0],))

    def _purge_property(self, table, field, obj, name, cursor):
        cursor.execute(
            f"SELECT DISTINCT id FROM {table} WHERE {field} = ? AND name = ? ORDER BY date",
            (obj, name),
        )
        cursor.execute(
            f"SELECT DISTINCT id FROM {table} WHERE {field} = ? AND name = ? ORDER BY date",
            (obj, name),
        )
        rows = cursor.fetchall()
        to_del = len(rows) - self.history_depth
        if to_del > 0:
            for row in rows[:to_del]:
                cursor.execute(f"DELETE FROM {table} WHERE id=?", (row[0],))

    def _get_device_host(self, name, cursor):
        name = replace_wildcard(name)
        cursor.execute(r"SELECT host FROM device WHERE name LIKE ? ESCAPE '\'", (name,))
        row = cursor.fetchone()
        if row is None:
            raise Exception("No host for device '" + name + "'")
        else:
            return row[0]

    def _send_starter_cmd(self, starter_dev_names):
        for name in starter_dev_names:
            pos = name.find(".")
            if pos != -1:
                name = name[0:pos]
            try:
                dev = tango.DeviceProxy(name)
                dev.UpdateServersInfo()
            except tango.DevFailed:
                pass

    # TANGO API

    def get_stored_procedure_release(self):
        return "release 1.8"

    @log_me
    @use_cursor
    def add_device(self, server_name, dev_info, klass_name, alias=None):
        self._info(
            "delete_attribute_alias(server_name=%s, dev_info=%s, klass_name=%s, alias=%s)",
            server_name,
            dev_info,
            klass_name,
            alias,
        )
        dev_name, (domain, family, member) = dev_info
        cursor = self.cursor
        # first delete the tuple (device,name) from the device table
        cursor.execute("DELETE FROM device WHERE name LIKE ?", (dev_name,))

        # then insert the new value for this tuple
        cursor.execute(
            "INSERT INTO device (name, alias, domain, family, member, exported, "
            "ior, host, server, pid, class, version, started, stopped) "
            'VALUES (?, ?, ?, ?, ?, 0, "nada", "nada", ?, 0, ?, "0", NULL, NULL)',
            (dev_name, alias, domain, family, member, server_name, klass_name),
        )

        # Check if a DServer device entry for the process already exists
        cursor.execute(
            'SELECT name FROM device WHERE server LIKE ? AND class LIKE "DServer"',
            (server_name,),
        )
        if cursor.fetchone() is None:
            dev_name = "dserver/" + server_name
            domain, family, member = dev_name.split("/", 2)
            cursor.execute(
                "INSERT INTO device (name, domain, family, member, exported, ior, "
                "host, server, pid, class, version, started, stopped) "
                'VALUES (?, ?, ?, ?, 0, "nada", "nada", ?, 0, "DServer", "0", NULL, NULL)',
                (dev_name, domain, family, member, server_name),
            )

    @log_me
    @use_cursor
    def delete_attribute_alias(self, alias):
        self._info("delete_attribute_alias(alias=%s)", alias)
        self.cursor.execute("DELETE FROM attribute_alias WHERE alias=?", (alias,))

    @log_me
    @use_cursor
    def delete_class_attribute(self, klass_name, attr_name):
        self.cursor.execute(
            "DELETE FROM property_attribute_class WHERE class LIKE ? AND "
            "attribute LIKE ?",
            (klass_name, attr_name),
        )

    @log_me
    @use_cursor
    def delete_class_attribute_property(self, klass_name, attr_name, prop_name):
        cursor = self.cursor

        # Is there something to delete ?
        cursor.execute(
            "SELECT count(*) FROM property_attribute_class WHERE class = ? "
            "AND attribute = ? AND name = ?",
            (klass_name, attr_name, prop_name),
        )
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        if cursor.fetchone()[0] > 0:
            # then delete property from the property_attribute_class table
            cursor.execute(
                "DELETE FROM property_attribute_class WHERE class = ? AND "
                "attribute = ? and name = ?",
                (klass_name, attr_name, prop_name),
            )
            # mark this property as deleted
            hist_id = self._get_id("class_attribute", cursor=cursor)
            cursor.execute(
                "INSERT INTO property_attribute_class_hist (date, class, attribute, "
                "name, id, count, value) VALUES "
                '(?, ?, ?, ?, ?, "0", "DELETED")',
                (now, klass_name, attr_name, prop_name, hist_id),
            )
            self._purge_att_property(
                "property_attribute_class_hist",
                "class",
                klass_name,
                attr_name,
                prop_name,
                cursor=cursor,
            )

    @log_me
    @use_cursor
    def delete_class_property(self, klass_name, prop_name):
        cursor = self.cursor

        prop_name = replace_wildcard(prop_name)
        # Is there something to delete ?
        cursor.execute(
            r"SELECT DISTINCT name FROM property_class WHERE class=? AND name LIKE ? ESCAPE '\'",
            (klass_name, prop_name),
        )

        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        for row in cursor.fetchall():
            # delete the tuple (device,name,count) from the property table
            name = row[0]
            cursor.execute(
                "DELETE FROM property_class WHERE class=? AND name=?",
                (klass_name, name),
            )
            # Mark this property as deleted
            hist_id = self._get_id("class", cursor=cursor)
            cursor.execute(
                "INSERT INTO property_class_hist (date, class, name, id, count, value)"
                + ' VALUES (?, ?, ?, ?, "0", "DELETED")',
                (now, klass_name, name, hist_id),
            )
            self._purge_property(
                "property_class_hist", "class", klass_name, name, cursor=cursor
            )

    @log_me
    @use_cursor
    def delete_device(self, dev_name):
        self._info("delete_device(dev_name=%s)", dev_name)
        cursor = self.cursor
        dev_name = replace_wildcard(dev_name)

        # delete the device from the device table
        cursor.execute(r"DELETE FROM device WHERE name LIKE ? ESCAPE '\'", (dev_name,))

        # delete device from the property_device table
        cursor.execute(
            r"DELETE FROM property_device WHERE device LIKE ? ESCAPE '\'", (dev_name,)
        )

        # delete device from the property_attribute_device table
        cursor.execute(
            r"DELETE FROM property_attribute_device WHERE device LIKE ? ESCAPE '\'",
            (dev_name,),
        )

    @log_me
    @use_cursor
    def delete_device_alias(self, dev_alias):
        self._info("delete_device_alias(dev_alias=%s)", dev_alias)
        self.cursor.execute("UPDATE device SET alias=NULL WHERE alias=?", (dev_alias,))

    @log_me
    @use_cursor
    def delete_device_attribute(self, dev_name, attr_name):
        dev_name = replace_wildcard(dev_name)
        self.cursor.execute(
            r"DELETE FROM property_attribute_device WHERE device LIKE ? ESCAPE '\'"
            + " AND attribute LIKE ?",
            (dev_name, attr_name),
        )

    @log_me
    @use_cursor
    def delete_device_attribute_property(self, dev_name, attr_name, prop_name):
        cursor = self.cursor
        # Is there something to delete ?
        cursor.execute(
            "SELECT count(*) FROM property_attribute_device WHERE device = ?"
            + " AND attribute = ? AND name = ?",
            (dev_name, attr_name, prop_name),
        )
        if cursor.fetchone()[0] > 0:
            # delete property from the property_attribute_device table
            cursor.execute(
                "DELETE FROM property_attribute_device WHERE device = ?"
                + " AND attribute = ? AND name = ?",
                (dev_name, attr_name, prop_name),
            )
            # Mark this property as deleted
            hist_id = self._get_id("device_attribute", cursor=cursor)
            # TODO I think this is incorrect; we need to insert all the rows
            # of the property here, to keep history intact
            cursor.execute(
                "INSERT INTO property_attribute_device_hist"
                + " (date, device, attribute, name, id, count, value)"
                + ' VALUES (datetime("now", "localtime"), ?, ?, ?, ?, "0", "DELETED")',
                (dev_name, attr_name, prop_name, hist_id),
            )
            self._purge_att_property(
                "property_attribute_device_hist",
                "device",
                dev_name,
                attr_name,
                prop_name,
                cursor=cursor,
            )

    @log_me
    @use_cursor
    def delete_device_property(self, dev_name, prop_name):
        cursor = self.cursor
        prop_name = replace_wildcard(prop_name)

        # Is there something to delete ?
        cursor.execute(
            "SELECT DISTINCT name, count FROM property_device WHERE device=?"
            + r" AND name LIKE ? ESCAPE '\'",
            (dev_name, prop_name),
        )
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        for row in cursor.fetchall():
            # delete the tuple (device,name,count) from the property table
            cursor.execute(
                r"DELETE FROM property_device WHERE device=? AND name LIKE ? ESCAPE '\'",
                (dev_name, prop_name),
            )
            # Mark this property as deleted
            hist_id = self._get_id("device", cursor=cursor)
            cursor.execute(
                "INSERT INTO property_device_hist (device, name, id, count, value, date) VALUES (?, ?, ?, ?, ?, ?)",
                (dev_name, row[0], hist_id, str(row[1]), "DELETED", now),
            )
            self._purge_property(
                "property_device_hist", "device", dev_name, row[0], cursor=cursor
            )

    @log_me
    @use_cursor
    def delete_property(self, obj_name, prop_name):
        cursor = self.cursor
        prop_name = replace_wildcard(prop_name)

        # Is there something to delete ?
        cursor.execute(
            r"SELECT DISTINCT name FROM property WHERE object= ? AND name LIKE ? ESCAPE '\'",
            (obj_name, prop_name),
        )
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        for row in cursor.fetchall():
            # delete the tuple (object,name,count) from the property table
            cursor.execute(
                r"DELETE FROM property WHERE object=? AND name LIKE ? ESCAPE '\'",
                (obj_name, prop_name),
            )
            # Mark this property as deleted
            hist_id = self._get_id("object", cursor=cursor)
            cursor.execute(
                "INSERT INTO property_hist (object, name, id, count, value, date) "
                'VALUES (?, ?, ?, "0", "DELETED", ?)',
                (obj_name, row[0], hist_id, now),
            )
            self._purge_property(
                "property_hist", "object", obj_name, row[0], cursor=cursor
            )

    @log_me
    @use_cursor
    def delete_server(self, server_instance):
        cursor = self.cursor
        server_instance = replace_wildcard(server_instance)

        previous_host = None
        # get host where running
        if self.fire_to_starter:
            adm_dev_name = "dserver/" + server_instance
            previous_host = self._get_device_host(adm_dev_name, cursor=cursor)

        # then delete the device from the device table
        cursor.execute(
            r"DELETE FROM device WHERE server LIKE ? ESCAPE '\'", (server_instance,)
        )

        # Update host's starter to update controlled servers list
        if self.fire_to_starter and previous_host:
            self._send_starter_cmd(previous_host)
            pass

    @log_me
    @use_cursor
    def delete_server_info(self, server_instance):
        self.cursor.execute("DELETE FROM server WHERE name=?", (server_instance,))

    @log_me
    @use_cursor
    def export_device(self, dev_name, IOR, host, pid, version):
        self._info(
            "export_device(dev_name=%s, host=%s, pid=%s, version=%s)",
            dev_name,
            host,
            pid,
            version,
        )
        self._info("export_device(IOR=%s)", IOR)
        cursor = self.cursor
        do_fire = False
        previous_host = None

        if self.fire_to_starter and dev_name[0:8] == "dserver/":
            # Get database server name
            tango_util = tango.Util.instance()
            db_serv = tango_util.get_ds_name()
            adm_dev_name = "dserver/" + db_serv.lower()
            if dev_name != adm_dev_name and dev_name[0:16] != "dserver/starter/":
                do_fire = True
                previous_host = self._get_device_host(dev_name, cursor=cursor)

        cursor.execute("SELECT server FROM device WHERE name = ?", (dev_name,))
        row = cursor.fetchone()
        if row is None:
            th_exc(
                DB_DeviceNotDefined,
                "device " + dev_name + " not defined in the database !",
                "DataBase::ExportDevice()",
            )
        server = row[0]

        # update the new value for this tuple
        cursor.execute(
            "UPDATE device SET exported=1, ior=?, host=?, pid=?, version=?, "
            'started=datetime("now", "localtime") WHERE name = ?',
            (IOR, host, pid, version, dev_name),
        )

        # update host name in server table
        cursor.execute("UPDATE server SET host=? WHERE name = ?", (host, server))

        if do_fire:
            hosts = []
            hosts.append(host)
            if (
                previous_host != ""
                and previous_host != "nada"
                and previous_host != host
            ):
                hosts.append(previous_host)
            self._send_starter_cmd(hosts)

    @log_me
    @use_cursor
    def export_event(self, event, IOR, host, pid, version):
        cursor = self.cursor
        cursor.execute(
            "INSERT INTO event (name,exported,ior,host,pid,version,started) "
            'VALUES (?, 1, ?, ?, ?, ?, datetime("now", "localtime"))',
            (event, IOR, host, pid, version),
        )

    @log_me
    @use_cursor
    def get_alias_device(self, dev_alias):
        cursor = self.cursor
        cursor.execute(
            r"SELECT name FROM device WHERE alias LIKE ? ESCAPE '\'", (dev_alias,)
        )
        row = cursor.fetchone()
        if row is None:
            th_exc(
                DB_DeviceNotDefined,
                "No device found for alias '" + dev_alias + "'",
                "DataBase::GetAliasDevice()",
            )
        return row[0]

    @log_me
    @use_cursor
    def get_attribute_alias(self, attr_alias):
        cursor = self.cursor
        cursor.execute(
            "SELECT name from attribute_alias WHERE alias LIKE ?", (attr_alias,)
        )
        row = cursor.fetchone()
        if row is None:
            th_exc(
                DB_SQLError,
                "No attribute found for alias '" + attr_alias + "'",
                "DataBase::GetAttributeAlias()",
            )
        return row[0]

    @log_me
    @use_cursor
    def get_attribute_alias_list(self, attr_alias):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT alias FROM attribute_alias WHERE alias LIKE ? ORDER BY attribute",
            (attr_alias,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_class_attribute_list(self, class_name, wildcard):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT attribute FROM property_attribute_class WHERE class=? and attribute like ?",
            (class_name, wildcard),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_class_attribute_property(self, class_name, attributes):
        cursor = self.cursor
        stmt = "SELECT name,value FROM property_attribute_class WHERE class=? AND attribute LIKE ?"
        result = [class_name, str(len(attributes))]
        for attribute in attributes:
            cursor.execute(stmt, (class_name, attribute))
            rows = cursor.fetchall()
            result.append(attribute)
            result.append(str(len(rows)))
            for row in rows:
                result.append(row[0])
                result.append(row[1])
        return result

    @log_me
    @use_cursor
    def get_class_attribute_property2(self, class_name, attributes):
        cursor = self.cursor
        stmt = "SELECT name,value FROM property_attribute_class WHERE class=? AND attribute LIKE ? ORDER BY name,count"
        # result = [class_name, str(len(attributes))]

        result = [class_name, str(len(attributes))]
        for attribute in attributes:
            cursor.execute(stmt, (class_name, attribute))
            result.append(attribute)
            nb_props = 0
            props = []
            rows = cursor.fetchall()
            for prop, grp in groupby(rows, itemgetter(0)):
                values = [value for _, value in grp]
                nb_props += 1
                props.append(prop)
                props.append(str(len(values)))
                props.extend(values)
            result.append(str(nb_props))
            result.extend(props)

        return result

    @log_me
    @use_cursor
    def get_class_attribute_property_hist(self, class_name, attribute, prop_name):
        stmt = """
        SELECT attribute, name, date, count, value FROM property_attribute_class_hist
        WHERE class = ? AND attribute LIKE ? AND name LIKE ? ORDER BY class, attribute, name, id, date ASC
        """

        result = []
        self.cursor.execute(stmt, (class_name, attribute, prop_name))
        for row in self.cursor.fetchall():
            result.extend(
                [
                    row["attribute"],
                    row["name"],
                    row["date"],
                    str(row["count"]),
                    row["value"],
                ]
            )
        return result

    @log_me
    @use_cursor
    def get_class_for_device(self, dev_name):
        return self._get_class_for_device(dev_name, self.cursor)

    def _get_class_for_device(self, dev_name, cursor):
        cursor.execute("SELECT DISTINCT class FROM device WHERE name = ?", (dev_name,))
        row = cursor.fetchone()
        if row is None:
            th_exc(
                DB_IncorrectArguments,
                "Class not found for " + dev_name,
                "Database.GetClassForDevice",
            )
        return row["class"]

    @log_me
    @use_cursor
    def get_class_inheritance_for_device(self, dev_name):
        cursor = self.cursor
        class_name = self._get_class_for_device(dev_name, cursor=cursor)
        props = self._get_class_property(class_name, ["InheritedFrom"], cursor=cursor)
        return [class_name] + props[4:]

    @log_me
    @use_cursor
    def get_class_list(self, server):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT class FROM device WHERE class LIKE ? ORDER BY class",
            (server,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_class_property(self, class_name, properties):
        return self._get_class_property(class_name, properties, self.cursor)

    def _get_class_property(self, class_name, properties, cursor):
        stmt = r"SELECT count, value, name FROM property_class WHERE class=? AND name LIKE ? ESCAPE '\' ORDER BY count"
        result = []
        result.append(class_name)
        result.append(str(len(properties)))
        for prop_name in properties:
            tmp_name = replace_wildcard(prop_name)
            cursor.execute(stmt, (class_name, tmp_name))
            rows = list(cursor.fetchall())
            result.append(prop_name)
            result.append(str(len(rows)))
            for row in rows:
                result.append(row[1])
        return result

    @log_me
    @use_cursor
    def get_class_property_hist(self, class_name, prop_name):
        cursor = self.cursor
        stmt = "SELECT DISTINCT id FROM property_class_hist WHERE class=? AND name LIKE ? ORDER by date ASC"

        result = []

        cursor.execute(stmt, (class_name, prop_name))

        for row in cursor.fetchall():
            idr = row[0]

            stmt = "SELECT strftime('%Y-%m-%d %H:%M:%S', date), name, value FROM property_class_hist WHERE id =? AND class =?"

            cursor.execute(stmt, (idr, class_name))

            rows = list(cursor.fetchall())
            date, name, _ = rows[0]
            result.append(name)
            result.append(date)
            result.append(str(len(rows)))
            for _, _, value in rows:
                result.append(value)

        return result

    @log_me
    @use_cursor
    def get_class_property_list(self, class_name):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT name FROM property_class WHERE class LIKE ? order by NAME",
            (class_name,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_alias(self, dev_name):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT alias FROM device WHERE name LIKE ?", (dev_name,)
        )
        try:
            (alias,) = cursor.fetchone()
        except (IndexError, TypeError):
            th_exc(
                DB_DeviceNotDefined,
                "No device called '" + dev_name + "'",
                "DataBase::GetDeviceAlias()",
            )
        return alias

    @log_me
    @use_cursor
    def get_device_alias_list(self, alias):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT alias FROM device WHERE alias LIKE ? ORDER BY alias",
            (alias,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_attribute_list(self, dev_name, attribute):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT  attribute FROM property_attribute_device WHERE device=?  AND attribute LIKE ? ORDER BY attribute",
            (
                dev_name,
                attribute,
            ),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_attribute_property(self, dev_name, attributes):
        cursor = self.cursor
        stmt = "SELECT name,value FROM property_attribute_device WHERE device=? AND attribute LIKE ?"
        result = [dev_name, str(len(attributes))]
        for attribute in attributes:
            cursor.execute(stmt, (dev_name, attribute))
            rows = cursor.fetchall()
            result.append(attribute)
            result.append(str(len(rows)))
            for row in rows:
                result.append(row[0])
                result.append(row[1])
        return result

    @log_me
    @use_cursor
    def get_device_attribute_property2(self, dev_name, attributes):
        cursor = self.cursor
        stmt = "SELECT name,value FROM property_attribute_device WHERE device=? AND attribute LIKE ? ORDER BY name,count"
        result = [dev_name, str(len(attributes))]

        for attribute in attributes:
            cursor.execute(stmt, (dev_name, attribute))
            result.append(attribute)
            nb_props = 0
            props = []
            rows = cursor.fetchall()
            for prop, grp in groupby(rows, itemgetter(0)):
                values = [value for _, value in grp]
                nb_props += 1
                props.append(prop)
                props.append(str(len(values)))
                props.extend(values)
            result.append(str(nb_props))
            result.extend(props)
        return result

    @log_me
    @use_cursor
    def get_device_attribute_property_hist(self, dev_name, attribute, prop_name):
        cursor = self.cursor
        stmt = "SELECT  DISTINCT id FROM property_attribute_device_hist WHERE device=? AND attribute LIKE ? AND name LIKE ? ORDER by date ASC"

        result = []

        cursor.execute(stmt, (dev_name, attribute, prop_name))

        for row in cursor.fetchall():
            idr = row[0]

            stmt = "SELECT attribute,name,strftime('%Y-%m-%d %H:%M:%S', date),value FROM property_attribute_device_hist WHERE id =? AND device =? ORDER BY count ASC"

            cursor.execute(stmt, (idr, dev_name))
            rows = cursor.fetchall()

            attribute, name, date = rows[0][:3]
            count = len(rows)
            result.extend([attribute, name, date, str(count)])
            for *_, value in rows:
                result.append(value)

        return result

    @log_me
    @use_cursor
    def get_device_class_list(self, server_name):
        cursor = self.cursor
        result = []
        cursor.execute(
            "SELECT name,class FROM device WHERE server =?  ORDER BY name",
            (server_name,),
        )
        for row in cursor.fetchall():
            result.append(row[0])
            result.append(row[1])

        return result

    @log_me
    @use_cursor
    def get_device_domain_list(self, wildcard):
        cursor = self.cursor
        cursor.execute(
            r"SELECT DISTINCT domain FROM device WHERE name LIKE ? ESCAPE '\' ORDER BY domain",
            (wildcard,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_exported_list(self, wildcard):
        cursor = self.cursor
        cursor.execute(
            r"SELECT DISTINCT name FROM device WHERE (name LIKE ? ESCAPE '\' OR alias LIKE ?  ESCAPE '\') AND exported=1 ORDER BY name",
            (wildcard, wildcard),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_family_list(self, wildcard):
        cursor = self.cursor
        cursor.execute(
            r"SELECT DISTINCT family FROM device WHERE name LIKE ? ESCAPE '\' ORDER BY family",
            (wildcard,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_info(self, dev_name):
        cursor = self.cursor
        cursor.execute(
            "SELECT exported,ior,version,pid,server,host,started,stopped,class FROM device WHERE name =?  or alias =?",
            (dev_name, dev_name),
        )
        result_long = []
        result_str = []
        for row in cursor.fetchall():
            if (row[4] is None) or (row[5] is None):
                th_exc(
                    DB_SQLError,
                    "Wrong info in database for device '" + dev_name + "'",
                    "DataBase::GetDeviceInfo()",
                )
            result_str.append(dev_name.lower())  # Lowercase for compatibility
            if row[1] is not None:
                result_str.append(str(row[1]))
            else:
                result_str.append("")
            result_str.append(str(row[2]))
            result_str.append(str(row[4]))
            result_str.append(str(row[5]))

            for i in range(0, 2):
                cursor.execute(
                    "SELECT strftime('%d-%m-%Y at %H:%M:%S', ?)", (row[6 + i],)
                )
                tmp_date = cursor.fetchone()[0]
                if tmp_date is None:
                    result_str.append("?")
                else:
                    result_str.append(str(tmp_date))
            result_str.append(str(row[8]))

            # TODO correct?
            # Exported
            try:
                result_long.append(None if row[0] == "nada" else int(row[0]))
            except ValueError:
                result_long.append(0)
            #
            try:
                result_long.append(int(row[3]))
            except ValueError:
                result_long.append(0)

        if not result_long:
            e0 = tango.DevError()
            e0.desc = f"device {dev_name} not defined in the database !"
            e0.origin = "DataBase::GetDeviceInfo()"
            e0.reason = "DB_DeviceNotDefined"
            e0.severity = tango.ErrSeverity.ERR
            raise tango.DevFailed(e0)

        result = (result_long, result_str)
        return result

    @log_me
    @use_cursor
    def get_device_list(self, server_name, class_name):
        cursor = self.cursor
        cursor.execute(
            r"SELECT DISTINCT name FROM device WHERE server LIKE ? ESCAPE '\' AND class LIKE ? ESCAPE '\' ORDER BY name",
            (server_name, class_name),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_wide_list(self, wildcard):
        cursor = self.cursor
        cursor.execute(
            r"SELECT DISTINCT name FROM device WHERE name LIKE ? ESCAPE '\' ORDER BY name",
            (wildcard,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_member_list(self, wildcard):
        cursor = self.cursor
        cursor.execute(
            r"SELECT DISTINCT  member FROM device WHERE name LIKE ? ESCAPE '\' ORDER BY member",
            (wildcard,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_property(self, dev_name, properties):
        cursor = self.cursor
        stmt = r"SELECT count,value,name FROM property_device WHERE device = ? AND name LIKE ? ESCAPE '\' ORDER BY count"
        result = []
        result.append(dev_name)
        result.append(str(len(properties)))
        for prop in properties:
            result.append(prop)
            tmp_name = replace_wildcard(prop)
            cursor.execute(stmt, (dev_name, tmp_name))
            rows = cursor.fetchall()
            result.append(str(len(rows)))
            for row in rows:
                result.append(row[1])
            if not rows:
                # TODO No idea what's the point of this but seems it's expected
                result.append(" ")
        return result

    @log_me
    @use_cursor
    def get_device_property_list(self, device_name, prop_filter):
        cursor = self.cursor
        cursor.execute(
            r"SELECT DISTINCT name FROM property_device WHERE device LIKE ? AND name LIKE ? ESCAPE '\' order by NAME",
            (device_name, prop_filter),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_device_property_hist(self, device_name, prop_name):
        cursor = self.cursor
        stmt1 = r"""
        SELECT DISTINCT id FROM property_device_hist
        WHERE device=? AND name LIKE ?  ESCAPE '\'
        ORDER by date ASC
        """
        result = []

        tmp_name = replace_wildcard(prop_name)

        stmt2 = r"""
        SELECT name, strftime('%Y-%m-%d %H:%M:%S', date), value
        FROM property_device_hist
        WHERE id = ? AND device = ? AND name LIKE ? ESCAPE '\'
        ORDER BY count ASC
        """

        cursor.execute(stmt1, (device_name, tmp_name))
        for (idr,) in cursor.fetchall():
            cursor.execute(stmt2, (idr, device_name, tmp_name))
            rows = list(cursor.fetchall())
            name, date, value = rows[0]
            result.extend([name, date, str(len(rows)), value])
            for row2 in rows[1:]:
                result.append(row2[2])
        return result

    @log_me
    @use_cursor
    def get_device_server_class_list(self, server_name):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT class FROM device WHERE server LIKE ? ORDER BY class",
            (server_name,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_exported_device_list_for_class(self, class_name):
        cursor = self.cursor
        cursor.execute(
            "SELECT  DISTINCT name FROM device WHERE class LIKE ? AND exported=1 ORDER BY name",
            (class_name,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_host_list(self, host_name):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT host FROM device WHERE host LIKE ?  ORDER BY host",
            (host_name,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_host_server_list(self, host_name):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT server FROM device WHERE host LIKE ?  ORDER BY server",
            (host_name,),
        )
        return [row[0] for row in cursor.fetchall()]

    def get_host_servers_info(self, host_name):
        servers = self.get_host_server_list(host_name)
        result = []
        for server in servers:
            result.append(server)
            info = self.get_server_info(server)
            result.append(info[2])
            result.append(info[3])
        return result

    def get_instance_name_list(self, server_name):
        server_name = server_name + "/%"
        server_list = self.get_server_list(server_name)
        result = []
        for server in server_list:
            names = server.split("/")
            result.append(names[1])
        return result

    @log_me
    @use_cursor
    def get_object_list(self, name):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT object FROM property WHERE object LIKE ?  ORDER BY object",
            (name,),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_property(self, object_name, properties):
        cursor = self.cursor
        result = []
        result.append(object_name)
        result.append(str(len(properties)))
        stmt = r"SELECT count,value,name FROM property WHERE object LIKE ?  AND name LIKE ? ESCAPE '\' ORDER BY count"
        for prop_name in properties:
            result.append(prop_name)
            prop_name = replace_wildcard(prop_name)
            cursor.execute(stmt, (object_name, prop_name))
            rows = cursor.fetchall()
            n_rows = len(rows)
            result.append(str(n_rows))
            if n_rows:
                for row in rows:
                    result.append(row[1])
            else:
                result.append(" ")
        return result

    @log_me
    @use_cursor
    def get_property_hist(self, object_name, prop_name):
        cursor = self.cursor
        result = []

        stmt = r"""
        SELECT  DISTINCT id FROM property_hist
        WHERE object=? AND name LIKE ? ESCAPE '\'
        ORDER by date ASC
        """
        prop_name = replace_wildcard(prop_name)
        cursor.execute(stmt, (object_name, prop_name))
        rows = cursor.fetchall()

        stmt2 = r"""
        SELECT name, strftime('%Y-%m-%d %H:%M:%S', date), value, count
        FROM property_hist
        WHERE id =? AND object = ? AND name LIKE ? ESCAPE '\'
        ORDER BY count ASC
        """
        for row in rows:
            idr = row[0]
            cursor.execute(stmt2, (idr, object_name, prop_name))
            rows2 = cursor.fetchall()
            name, date, value, count = rows2[0]
            result.extend([name, date, str(len(rows2)), value])
            for _, _, value, _ in rows2[1:]:
                result.append(value)
        return result

    @log_me
    @use_cursor
    def get_property_list(self, object_name, wildcard):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT name FROM property WHERE object LIKE ? AND name LIKE ? ORDER BY name",
            (object_name, wildcard),
        )
        return [row[0] for row in cursor.fetchall()]

    @log_me
    @use_cursor
    def get_server_info(self, server_name):
        cursor = self.cursor
        cursor.execute(
            "SELECT host, mode, level FROM server WHERE name = ?", (server_name,)
        )
        result = []
        result.append(server_name)
        row = cursor.fetchone()
        if row is None:
            # TODO would it not make more sense to throw an error if no server was found,
            # instead of returning made up data?
            result.append(" ")
            result.append(" ")
            result.append(" ")
        else:
            result.append(row[0])
            result.append(str(row[1]))
            result.append(str(row[2]))

        return result

    @log_me
    @use_cursor
    def _get_server_list(self, wildcard):
        cursor = self.cursor
        cursor.execute(
            "SELECT DISTINCT server FROM device WHERE server LIKE ? ORDER BY server",
            (wildcard,),
        )
        return [row[0] for row in cursor.fetchall()]

    def get_server_list(self, wildcard):
        result = []
        server_list = self._get_server_list(wildcard)
        return server_list
        for server in server_list:
            found = 0
            server_name = server.split("/")[0]
            for res in result:
                if server_name.lower() == res.lower():
                    found = 1
            if not found:
                result.append(server_name)
        return result

    def get_server_name_list(self, wildcard):
        result = []
        server_list = self._get_server_list(wildcard)
        for server in server_list:
            found = 0
            server_name = server.split("/")[0]
            for res in result:
                if server_name.lower() == res.lower():
                    found = 1
            if not found:
                result.append(server_name)
        return result

    @log_me
    @use_cursor
    def import_device(self, dev_name):
        cursor = self.cursor
        result_long = []
        result_str = []
        # Search first by server name and if nothing found by alias
        # Using OR takes much more time
        cursor.execute(
            "SELECT exported,ior,version,pid,server,host,class FROM device WHERE name =? COLLATE NOCASE",
            (dev_name,),
        )
        rows = cursor.fetchall()
        if len(rows) == 0:
            cursor.execute(
                "SELECT exported,ior,version,pid,server,host,class FROM device WHERE alias =? COLLATE NOCASE",
                (dev_name,),
            )
            rows = cursor.fetchall()
            if len(rows) == 0:
                th_exc(
                    DB_DeviceNotDefined,
                    "device " + dev_name + " not defined in the database !",
                    "DataBase::ImportDevice()",
                )
        for row in rows:
            result_str.append(dev_name)
            result_str.append(row[1])
            result_str.append(row[2] if row[2] != "nada" else "0")
            result_str.append(row[4])
            result_str.append(row[5])
            result_str.append(row[6])

            if row[0] != "nada":
                result_long.append(row[0])
            else:
                result_long.append(0)
            if row[3] != "nada":
                result_long.append(row[3])
            else:
                result_long.append(0)

        result = (result_long, result_str)
        return result

    @log_me
    @use_cursor
    def import_event(self, event_name):
        cursor = self.cursor
        result_long = []
        result_str = []
        cursor.execute(
            "SELECT exported,ior,version,pid,host FROM event WHERE name =?",
            (event_name,),
        )
        rows = cursor.fetchall()
        if len(rows) == 0:
            th_exc(
                DB_DeviceNotDefined,
                "event " + event_name + " not defined in the database !",
                "DataBase::ImportEvent()",
            )
        for row in rows:
            result_str.append(event_name)
            result_str.append(row[1])
            result_str.append(row[2])
            result_str.append(row[4])
            exported = -1
            if row[0] is not None and row[0] != "nada":
                exported = row[0]
            result_long.append(exported)
            pid = 0
            if row[3] is not None and row[2] != "nada":
                pid = row[3]
            result_long.append(pid)
        result = (result_long, result_str)
        return result

    @log_me
    @use_cursor
    def info(self):
        cursor = self.cursor
        result = []
        # db name
        info_str = "TANGO Database " + self.db_name
        result.append(info_str)
        # new line
        result.append(" ")
        # get start time of database
        cursor.execute("SELECT started FROM device WHERE name =?", (self.dev_name,))
        row = cursor.fetchone()
        info_str = "Running since ..." + str(row[0])
        result.append(info_str)
        # new line
        result.append(" ")
        # get number of devices defined
        cursor.execute("SELECT COUNT(*) FROM device")
        row = cursor.fetchone()
        info_str = "Devices defined = " + str(row[0])
        result.append(info_str)
        # get number of devices exported
        cursor.execute("SELECT COUNT(*) FROM device WHERE exported = 1")
        row = cursor.fetchone()
        info_str = "Devices exported = " + str(row[0])
        result.append(info_str)
        # get number of device servers defined
        cursor.execute('SELECT COUNT(*) FROM device WHERE class = "DServer" ')
        row = cursor.fetchone()
        info_str = "Device servers defined = " + str(row[0])
        result.append(info_str)
        # get number of device servers exported
        cursor.execute(
            'SELECT COUNT(*) FROM device WHERE class = "DServer"  AND exported = 1'
        )
        row = cursor.fetchone()
        info_str = "Device servers exported = " + str(row[0])
        result.append(info_str)
        # new line
        result.append(" ")
        # get number of device properties
        cursor.execute("SELECT COUNT(*) FROM property_device")
        row = cursor.fetchone()
        info_str = "Device properties defined = " + str(row[0])
        cursor.execute("SELECT COUNT(*) FROM property_device_hist")
        row = cursor.fetchone()
        info_str = info_str + " [History lgth = " + str(row[0]) + "]"
        result.append(info_str)
        # get number of class properties
        cursor.execute("SELECT COUNT(*) FROM property_class")
        row = cursor.fetchone()
        info_str = "Class properties defined = " + str(row[0])
        cursor.execute("SELECT COUNT(*) FROM property_class_hist")
        row = cursor.fetchone()
        info_str = info_str + " [History lgth = " + str(row[0]) + "]"
        result.append(info_str)
        # get number of device attribute properties
        cursor.execute("SELECT COUNT(*) FROM property_attribute_device")
        row = cursor.fetchone()
        info_str = "Device attribute properties defined = " + str(row[0])
        cursor.execute("SELECT COUNT(*) FROM property_attribute_device_hist")
        row = cursor.fetchone()
        info_str = info_str + " [History lgth = " + str(row[0]) + "]"
        result.append(info_str)
        # get number of class attribute properties
        cursor.execute("SELECT COUNT(*) FROM property_attribute_class")
        row = cursor.fetchone()
        info_str = "Class attribute properties defined = " + str(row[0])
        cursor.execute("SELECT COUNT(*) FROM property_attribute_class_hist")
        row = cursor.fetchone()
        info_str = info_str + " [History lgth = " + str(row[0]) + "]"
        result.append(info_str)
        # get number of object properties
        cursor.execute("SELECT COUNT(*) FROM property")
        row = cursor.fetchone()
        info_str = "Object properties defined = " + str(row[0])
        cursor.execute("SELECT COUNT(*) FROM property_hist")
        row = cursor.fetchone()
        info_str = info_str + " [History lgth = " + str(row[0]) + "]"
        result.append(info_str)

        return result

    @log_me
    @use_cursor
    def put_attribute_alias(self, attribute_name, attribute_alias):
        cursor = self.cursor
        attribute_name = attribute_name.lower()
        # first check if this alias exists
        cursor.execute(
            "SELECT alias from attribute_alias WHERE alias=? AND name <> ? ",
            (attribute_alias, attribute_name),
        )
        rows = cursor.fetchall()
        if len(rows) > 0:
            self._warn("DataBase::DbPutAttributeAlias(): this alias exists already ")
            th_exc(
                DB_SQLError,
                "alias " + attribute_alias + " already exists !",
                "DataBase::DbPutAttributeAlias()",
            )
        tmp_names = attribute_name.split("/")
        if len(tmp_names) != 4:
            self._warn(
                "DataBase::DbPutAttributeAlias(): attribute name has bad syntax, must have 3 / in it"
            )
            th_exc(
                DB_SQLError,
                "attribute name "
                + attribute_name
                + " has bad syntax, must have 3 / in it",
                "DataBase::DbPutAttributeAlias()",
            )
        # first delete the current entry (if any)
        cursor.execute("DELETE FROM attribute_alias WHERE name=?", (attribute_name,))
        # update the new value for this tuple
        tmp_device = tmp_names[0] + "/" + tmp_names[1] + "/" + tmp_names[2]
        tmp_attribute = tmp_names[3]
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        cursor.execute(
            "INSERT INTO attribute_alias (alias, name, device, attribute, updated, accessed) VALUES(?, ?, ?, ?, ?, ?)",
            (attribute_alias, attribute_name, tmp_device, tmp_attribute, now, now),
        )

    @log_me
    @use_cursor
    def put_class_attribute_property(self, class_name, nb_attributes, attr_prop_list):
        cursor = self.cursor
        k = 0
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        for _i in range(0, nb_attributes):
            tmp_attribute = attr_prop_list[k]
            nb_properties = int(attr_prop_list[k + 1])
            for j in range(k + 2, k + nb_properties * 2 + 2, 2):
                tmp_name = attr_prop_list[j]
                tmp_value = attr_prop_list[j + 1]
                # first delete the tuple (device,name,count) from the property table
                cursor.execute(
                    "DELETE FROM property_attribute_class WHERE class LIKE ? AND attribute LIKE ? AND name LIKE ?",
                    (class_name, tmp_attribute, tmp_name),
                )
                # then insert the new value for this tuple
                cursor.execute(
                    "INSERT INTO property_attribute_class (class, attribute, name, count, value, updated, accessed) VALUES (?, ?, ?, '1', ?, ?, ?)",
                    (class_name, tmp_attribute, tmp_name, tmp_value, now, now),
                )
                # then insert the new value into the history table
                hist_id = self._get_id("class_attribute", cursor=cursor)
                cursor.execute(
                    "INSERT INTO property_attribute_class_hist (class, attribute, name, id, count, value, date) VALUES (?, ?, ?, ?, '1', ?, ?)",
                    (class_name, tmp_attribute, tmp_name, hist_id, tmp_value, now),
                )

                self._purge_att_property(
                    "property_attribute_class_hist",
                    "class",
                    class_name,
                    tmp_attribute,
                    tmp_name,
                    cursor=cursor,
                )
            k = k + nb_properties * 2 + 2

    @log_me
    @use_cursor
    def put_class_attribute_property2(self, class_name, nb_attributes, attr_prop_list):
        cursor = self.cursor
        k = 0
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        for _i in range(0, nb_attributes):
            tmp_attribute = attr_prop_list[k]
            nb_properties = int(attr_prop_list[k + 1])
            for _jj in range(0, nb_properties, 1):
                j = k + 2
                tmp_name = attr_prop_list[j]
                # first delete the tuple (device,name,count) from the property table
                cursor.execute(
                    "DELETE FROM property_attribute_class WHERE class LIKE ? AND attribute LIKE ? AND name LIKE ?",
                    (class_name, tmp_attribute, tmp_name),
                )
                n_rows = int(attr_prop_list[j + 1])
                tmp_count = 0
                for _l in range(j + 1, j + n_rows + 1, 1):
                    tmp_value = attr_prop_list[_l + 1]
                    tmp_count = tmp_count + 1
                    # then insert the new value for this tuple
                    cursor.execute(
                        "INSERT INTO property_attribute_class (class, attribute, name, count, value, updated, accessed) VALUES (? ,?,?,?,?,?,?)",
                        (
                            class_name,
                            tmp_attribute,
                            tmp_name,
                            str(tmp_count),
                            tmp_value,
                            now,
                            now,
                        ),
                    )
                    # then insert the new value into the history table
                    hist_id = self._get_id("class_attribute", cursor=cursor)
                    cursor.execute(
                        "INSERT INTO property_attribute_class_hist (date, class, attribute, name, id, count, value) VALUES (?,?,?,?,?,?,?)",
                        (
                            now,
                            class_name,
                            tmp_attribute,
                            tmp_name,
                            hist_id,
                            str(tmp_count),
                            tmp_value,
                        ),
                    )

                    self._purge_att_property(
                        "property_attribute_class_hist",
                        "class",
                        class_name,
                        tmp_attribute,
                        tmp_name,
                        cursor=cursor,
                    )
                k = k + n_rows + 2
            k = k + 2

    @log_me
    @use_cursor
    def put_class_property(self, class_name, nb_properties, prop_list):
        cursor = self.cursor
        hist_id = self._get_id(
            "device", cursor=cursor
        )  # Single id for the whole operation
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        property_values = []
        lines = iter(prop_list)
        while len(property_values) < nb_properties:
            name = next(lines)
            number_of_lines = int(next(lines))
            value = [next(lines) for i in range(number_of_lines)]
            property_values.append((name, value))
            # Delete current property
            cursor.execute(
                "DELETE FROM property_class WHERE class LIKE ? AND name LIKE ?",
                (class_name, name),
            )
            # Insert into property table
            cursor.executemany(
                "INSERT INTO property_class (class, name, count, value, updated, accessed) VALUES (?, ?, ?, ?, ?, ?)",
                [
                    (class_name, name, str(i), line, now, now)
                    for i, line in enumerate(value, start=1)
                ],
            )
            # Insert into property history table
            cursor.executemany(
                "INSERT INTO property_class_hist (class, name, id, count, value, date) VALUES (?, ?, ?, ?, ?, ?)",
                [
                    (class_name, name, hist_id, str(i), line, now)
                    for i, line in enumerate(value, start=1)
                ],
            )
            # Make sure history is limited
            self._purge_property(
                "property_class_hist", "class", class_name, name, cursor=cursor
            )

    @log_me
    @use_cursor
    def put_device_alias(self, device_name, device_alias):
        cursor = self.cursor
        device_name = device_name.lower()
        # first check if this alias exists
        cursor.execute(
            "SELECT alias from device WHERE alias=? AND name <>?",
            (device_alias, device_name),
        )
        rows = cursor.fetchall()
        if len(rows) > 0:
            self._warn("DataBase::DbPutDeviceAlias(): this alias exists already ")
            th_exc(
                DB_SQLError,
                "alias " + device_alias + " already exists !",
                "DataBase::DbPutDeviceAlias()",
            )
        # update the new value for this tuple
        cursor.execute(
            "UPDATE device SET alias=? ,started=datetime('now', 'localtime') where name LIKE ?",
            (device_alias, device_name),
        )

    @log_me
    @use_cursor
    def put_device_attribute_property(self, device_name, nb_attributes, attr_prop_list):
        # TODO implement this using put_device_attribute_property2?
        cursor = self.cursor
        k = 0
        for _i in range(0, nb_attributes):
            tmp_attribute = attr_prop_list[k]
            nb_properties = int(attr_prop_list[k + 1])
            now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

            for j in range(k + 2, k + nb_properties * 2 + 2, 2):
                tmp_name = attr_prop_list[j]
                tmp_value = attr_prop_list[j + 1]
                # first delete the tuple (device,name,count) from the property table
                cursor.execute(
                    "DELETE FROM property_attribute_device WHERE device LIKE ? AND attribute LIKE ? AND name LIKE ?",
                    (device_name, tmp_attribute, tmp_name),
                )
                # then insert the new value for this tuple
                cursor.execute(
                    "INSERT INTO property_attribute_device (device, attribute, name, count, value, updated, accessed) VALUES (?, ?, ?, '1', ?, ?, ?)",
                    (device_name, tmp_attribute, tmp_name, tmp_value, now, now),
                )
                # then insert the new value into the history table
                hist_id = self._get_id("device_attribute", cursor=cursor)
                cursor.execute(
                    "INSERT INTO property_attribute_device_hist (date,device,attribute,name,id,count,value) VALUES (?,?,?,?,?,'1',?)",
                    (now, device_name, tmp_attribute, tmp_name, hist_id, tmp_value),
                )

                self._purge_att_property(
                    "property_attribute_device_hist",
                    "device",
                    device_name,
                    tmp_attribute,
                    tmp_name,
                    cursor=cursor,
                )
            k = k + nb_properties * 2 + 2

    @log_me
    @use_cursor
    def put_device_attribute_property2(
        self, device_name, nb_attributes, attr_prop_list
    ):
        cursor = self.cursor

        # Group the data
        attr_prop = {}
        items = iter(attr_prop_list)
        while len(attr_prop) < nb_attributes:
            attrname = next(items)
            attr = attr_prop[attrname] = {}
            n_props = int(next(items))
            for _i in range(n_props):
                propname = next(items)
                prop = attr[propname] = []
                n_lines = int(next(items))
                for _j in range(n_lines):
                    line = next(items)
                    prop.append(line)

        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        hist_id = self._get_id("device_attribute", cursor=cursor)

        for attr_name, props in attr_prop.items():
            for prop_name, lines in props.items():
                cursor.execute(
                    "DELETE FROM property_attribute_device WHERE device LIKE ? AND attribute LIKE ? AND name LIKE ?",
                    (device_name, attr_name, prop_name),
                )
                for count, value in enumerate(lines, start=1):
                    cursor.execute(
                        "INSERT INTO property_attribute_device (device,attribute,name,count,value,updated,accessed) VALUES (?, ?, ?, ?, ?, ?, ?)",
                        (
                            device_name,
                            attr_name,
                            prop_name,
                            str(count),
                            value,
                            now,
                            now,
                        ),
                    )
                    cursor.execute(
                        "INSERT INTO property_attribute_device_hist (device,attribute,name,id,count,value,date) VALUES (?, ?, ?, ?, ?, ?, ?)",
                        (
                            device_name,
                            attr_name,
                            prop_name,
                            hist_id,
                            str(count),
                            value,
                            now,
                        ),
                    )
                self._purge_att_property(
                    "property_attribute_device_hist",
                    "device",
                    device_name,
                    attr_name,
                    prop_name,
                    cursor=cursor,
                )

    @log_me
    @use_cursor
    def put_device_property(self, device_name, nb_properties, prop_list):
        cursor = self.cursor
        hist_id = self._get_id(
            "device", cursor=cursor
        )  # Single id for the whole operation
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # attr_prop_list consists of nb_properties number of:
        # - property name, e.g. "MyProperty"
        # - number of lines in the value, e.g. "3"
        # - that many arbitrary strings, here 3
        #
        # Example (nb_properties = 2)
        # - MyProperty
        # - 3
        # - line1
        # - line2
        # - line3
        # - MyOtherProperty
        # - 1
        # - something
        #
        # TODO verify this structure, and raise a helpful error
        property_values = []
        lines = iter(prop_list)
        while len(property_values) < nb_properties:
            name = next(lines)
            number_of_lines = int(next(lines))
            value = [next(lines) for i in range(number_of_lines)]
            property_values.append((name, value))
            # Delete current property
            cursor.execute(
                "DELETE FROM property_device WHERE device LIKE ? AND name LIKE ?",
                (device_name, name),
            )
            # Insert into property table
            cursor.executemany(
                "INSERT INTO property_device (device, name, count, value, updated, accessed) VALUES (?, ?, ?, ?, ?, ?)",
                [
                    (device_name, name, str(i), line, now, now)
                    for i, line in enumerate(value, start=1)
                ],
            )
            # Insert into property history table
            cursor.executemany(
                "INSERT INTO property_device_hist (device, name, id, count, value, date) VALUES (?, ?, ?, ?, ?, ?)",
                [
                    (device_name, name, hist_id, str(i), line, now)
                    for i, line in enumerate(value, start=1)
                ],
            )
            self._purge_property(
                "property_device_hist", "device", device_name, name, cursor=cursor
            )

    @log_me
    @use_cursor
    def put_property(self, object_name, nb_properties, prop_list):
        cursor = self.cursor
        hist_id = self._get_id("object", cursor=cursor)
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        property_values = []
        lines = iter(prop_list)
        while len(property_values) < nb_properties:
            name = next(lines)
            number_of_lines = int(next(lines))
            value = [next(lines) for i in range(number_of_lines)]
            property_values.append((name, value))
            # first delete the property from the property table
            cursor.execute(
                "DELETE FROM property WHERE object = ? AND name = ?",
                (object_name, name),
            )
            # Insert into property table
            cursor.executemany(
                "INSERT INTO property (object, name, count, value, updated, accessed) VALUES (?, ?, ?, ?, ?, ?)",
                [
                    (object_name, name, str(i), line, now, now)
                    for i, line in enumerate(value, start=1)
                ],
            )
            # Insert into property history table
            cursor.executemany(
                "INSERT INTO property_hist (object, name, id, count, value, date) VALUES (?, ?, ?, ?, ?, ?)",
                [
                    (object_name, name, hist_id, str(i), line, now)
                    for i, line in enumerate(value, start=1)
                ],
            )
            self._purge_property(
                "property_hist", "object", object_name, name, cursor=cursor
            )

    @log_me
    @use_cursor
    def put_server_info(self, tmp_server, tmp_host, tmp_mode, tmp_level, tmp_extra):
        cursor = self.cursor
        # If it is an empty host name -> get previous host where running
        previous_host = ""
        if self.fire_to_starter and tmp_host == "":
            adm_dev_name = "dserver/" + tmp_server
            previous_host = self._get_device_host(adm_dev_name, cursor=cursor)
        # first delete the server from the server table
        cursor.execute("DELETE FROM server WHERE name=?", (tmp_server,))
        # insert the new info for this server
        cursor.execute(
            "INSERT INTO server (name, host, mode, level) VALUES (?, ?, ?, ?)",
            (tmp_server, tmp_host, tmp_mode, tmp_level),
        )
        #  Update host's starter to update controlled servers list
        if self.fire_to_starter:
            hosts = []
            if previous_host == "":
                hosts.append(tmp_host)
            else:
                hosts.append(previous_host)
            self._send_starter_cmd(hosts)

    @log_me
    @use_cursor
    def uexport_device(self, dev_name):
        cursor = self.cursor
        # self._info("un-export device(dev_name=%s)", dev_name)
        cursor.execute(
            "UPDATE device SET exported=0,stopped=datetime('now', 'localtime') WHERE name LIKE ?",
            (dev_name,),
        )

    @log_me
    @use_cursor
    def uexport_event(self, event_name):
        cursor = self.cursor
        self._info("un-export event (event_name=%s)", event_name)
        cursor.execute(
            "UPDATE event SET exported=0,stopped=datetime('now', 'localtime') WHERE name LIKE ?",
            (event_name,),
        )

    @log_me
    @use_cursor
    def uexport_server(self, server_name):
        cursor = self.cursor
        # self._info(f"un-export all devices from server {server_name}")
        cursor.execute(
            "UPDATE device SET exported=0,stopped=datetime('now', 'localtime') WHERE server LIKE ?",
            (server_name,),
        )

    @log_me
    @use_cursor
    def delete_all_device_attribute_property(self, dev_name, attr_list):
        cursor = self.cursor
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        for attr_name in attr_list:
            self._info(
                "_delete_all_device_attribute_property(): delete device %s attribute %s property(ies) from database",
                dev_name,
                attr_name,
            )
            # Is there something to delete ?
            cursor.execute(
                "SELECT DISTINCT name FROM property_attribute_device WHERE device =? AND attribute = ?",
                (dev_name, attr_name),
            )
            rows = cursor.fetchall()
            if len(rows) != 0:
                cursor.execute(
                    "DELETE FROM property_attribute_device WHERE device = ? AND attribute = ?",
                    (dev_name, attr_name),
                )
            # Mark this property as deleted
            for row in rows:
                hist_id = self._get_id("device_attribute", cursor=cursor)
                cursor.execute(
                    "INSERT INTO property_attribute_device_hist (date,device,attribute,name,id,count,value) VALUES (?,?,?,?,?,'0','DELETED')",
                    (now, dev_name, attr_name, row[0], hist_id),
                )
                self._purge_att_property(
                    "property_attribute_device_hist",
                    "device",
                    dev_name,
                    attr_name,
                    row[0],
                    cursor=cursor,
                )

    @log_me
    @use_cursor
    def my_sql_select(self, cmd):
        cursor = self.cursor
        cursor.execute(cmd)
        result_long = []
        result_str = []
        rows = list(cursor.fetchall())
        for row in rows:
            if row is None:
                result_str.append("")
                result_long.append(0)
            else:
                for field in row:
                    if field is not None:
                        result_str.append(str(field))
                        result_long.append(1)
                    else:
                        result_str.append("")
                        result_long.append(0)
        nb_fields = len(rows[0]) if rows else 1

        result_long.append(len(rows))
        result_long.append(nb_fields)

        result = (result_long, result_str)
        return result

    @log_me
    @use_cursor
    def get_csdb_server_list(self):
        cursor = self.cursor

        cursor.execute(
            "SELECT DISTINCT ior FROM device WHERE exported=1 AND domain='sys' AND family='database'"
        )
        results = []
        for row in cursor.fetchall():
            info = parse_ior(row[0])
            results.append(f"{info.host.decode()}:{info.port}")
        return results

    @log_me
    @use_cursor
    def get_attribute_alias2(self, attr_name):
        cursor = self.cursor
        cursor.execute(
            "SELECT alias from attribute_alias WHERE name LIKE ? ", (attr_name,)
        )
        rows = list(cursor.fetchall())
        if rows:
            # TODO What if we get more than one match? Probably not possible?
            return rows[0][0]
        th_exc(
            DB_SQLError,
            f"No attribute found for attribute '{attr_name}'",
            "DataBase::db_get_attribute_alias2()",
        )

    @log_me
    @use_cursor
    def get_alias_attribute(self, alias):
        cursor = self.cursor
        cursor.execute("SELECT name from attribute_alias WHERE alias LIKE ? ", (alias,))
        rows = cursor.fetchall()
        if rows:
            # TODO What if we get more than one match? Probably not possible?
            return rows[0][0]
        th_exc(
            DB_SQLError,
            f"No device found for alias '{alias}' ",
            "DataBase::db_get_alias_attribute()",
        )

    @log_me
    @use_cursor
    def rename_server(self, old_name, new_name):
        cursor = self.cursor
        # Check that the new name is not already used
        new_adm_name = "dserver/" + new_name
        cursor.execute("SELECT name from device WHERE name = ? ", (new_adm_name,))
        rows = cursor.fetchall()
        if len(rows) != 0:
            th_exc(
                DB_SQLError,
                f"Device server process name {new_name} is already used !",
                "DataBase::DbRenameServer()",
            )

        # get host where running
        previous_host = ""
        if self.fire_to_starter:
            try:
                adm_dev = "dserver/" + old_name
                previous_host = self._get_device_host(adm_dev, cursor=cursor)
            except Exception:
                th_exc(
                    DB_IncorrectServerName,
                    "Server " + old_name + "not defined in database!",
                    "DataBase::DbRenameServer()",
                )
        # Change ds exec name. This means
        #  1 - Update the device's server column
        #  2 - Change the ds admin device name
        #  3 - Change admin device property (if any)
        #  4 - Change admin device attribute property (if any)

        old_adm_name = "dserver/" + old_name
        tmp_new = new_name.split("/")
        new_exec = tmp_new[0]
        new_inst = tmp_new[1]
        new_adm_name = "dserver/" + new_name

        cursor.execute(
            "UPDATE device SET name =?, family =?, mamber =? WHERE name =?",
            (new_adm_name, new_exec, new_inst, old_adm_name),
        )

        cursor.execute(
            "UPDATE property_device set device=? WHERE device=?",
            (new_adm_name, old_adm_name),
        )

        cursor.execute(
            "UPDATE property_attribute_device set device=? WHERE device=?",
            (new_adm_name, old_adm_name),
        )

        #  Update host's starter to update controlled servers list
        if self.fire_to_starter:
            hosts = []
            if previous_host != "":
                hosts.append(previous_host)
            self._send_starter_cmd(hosts)


class Sqlite3Database(SqlDatabase):
    DB_API_NAME = "sqlite3"


def main():
    db = Sqlite3Database("2")
    db.add_device("MyServer/my1", ("a/b/c", ("a", "b", "c")), "MyClass")
    db.close_db()


def get_db(personal_name="2"):
    #    return Executor.submit(Sqlite3Database).result()
    return Sqlite3Database(personal_name)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import sys
import time
from functools import partial

import numpy as np
import pytest

from tango import (
    DevState,
    AttributeProxy,
    GreenMode,
    DevFailed,
    EventType,
    AttrWriteType,
)
from tango.constants import DefaultPollRingDepth
from tango.asyncio import AttributeProxy as asyncio_AttributeProxy
from tango.gevent import AttributeProxy as gevent_AttributeProxy
from tango.futures import AttributeProxy as futures_AttributeProxy

from tango.server import Device, attribute, command

from tango.test_utils import assert_close, DeviceTestContext, MultiDeviceTestContext
from tango.utils import EventCallback, AsyncEventCallback


TEST_VALUES = {
    "scalar_int": (2, 3, 4, 5, 6),
    "spectrum_str": (["c", "d"], ["e", "f"], ["g", "h"], ["i", "j"], ["k", "l"]),
    "image_float": (
        [[15.0, 16.0], [17.0, 18.0]],
        [[19.0, 20.0], [21.0, 22.0]],
        [[23.0, 24.0], [25.0, 26.0]],
        [[27.0, 28.0], [29.0, 30.0]],
        [[31.0, 32.0], [33.0, 34.0]],
    ),
}

ATTRIBUTES_TO_TEST = list(TEST_VALUES.keys())

attribute_proxy_map = {
    GreenMode.Synchronous: AttributeProxy,
    GreenMode.Futures: futures_AttributeProxy,
    GreenMode.Asyncio: partial(asyncio_AttributeProxy, wait=True),
    GreenMode.Gevent: gevent_AttributeProxy,
}

# Tests


class EasyEchoDevice(Device):

    scalar_int_value = 1
    spectrum_str_value = ["a", "b"]
    image_float_value = [[1.0, 2.0], [3.0, 4.0]]

    def init_device(self):
        self.set_state(DevState.ON)

    @attribute(access=AttrWriteType.READ_WRITE)
    def scalar_int(self) -> int:
        return self.scalar_int_value

    @scalar_int.setter
    def set_scalar_int(self, new_value):
        self.scalar_int_value = new_value

    @attribute(access=AttrWriteType.READ_WRITE)
    def spectrum_str(self) -> tuple[str, str]:
        return self.spectrum_str_value

    @spectrum_str.setter
    def set_spectrum_str(self, new_value):
        self.spectrum_str_value = new_value

    @attribute(access=AttrWriteType.READ_WRITE)
    def image_float(self) -> tuple[tuple[float, float], tuple[float, float]]:
        return self.image_float_value

    @image_float.setter
    def set_image_float(self, new_value):
        self.image_float_value = new_value


devices_info = ({"class": EasyEchoDevice, "devices": [{"name": "test/dev/main"}]},)


@pytest.fixture(params=ATTRIBUTES_TO_TEST)
def attribute_proxy(request):
    with MultiDeviceTestContext(devices_info=devices_info):
        proxy = AttributeProxy(f"test/dev/main/{request.param}")
        assert proxy.__repr__() == proxy.__str__() == f"AttributeProxy({request.param})"
        yield proxy


def test_ping(attribute_proxy):
    duration = attribute_proxy.ping(wait=True)
    assert isinstance(duration, int)


def test_state_status(attribute_proxy):
    state = attribute_proxy.state(wait=True)
    assert isinstance(state, DevState)

    status = attribute_proxy.status(wait=True)
    assert status == f"The device is in {state} state."


def test_read_write_attribute(attribute_proxy):
    values = TEST_VALUES[attribute_proxy.name()]
    attribute_proxy.write(values[0], wait=True)
    assert_close(attribute_proxy.read(wait=True).value, values[0])
    assert_close(attribute_proxy.write_read(values[1], wait=True).value, values[1])


def test_attribute_poll(attribute_proxy):
    poll_period = 0.1  # sec
    values = TEST_VALUES[attribute_proxy.name()]
    initial_value = attribute_proxy.read().value
    t_start = time.time()

    _assert_polling_can_be_started(attribute_proxy, poll_period)
    history = _write_values_and_read_via_polling(attribute_proxy, poll_period, values)
    _assert_polling_can_be_stopped(attribute_proxy)
    _assert_reading_times_increase_monotonically(history, t_start)
    _assert_reading_values_valid(history, initial_value, values)


def _assert_polling_can_be_started(attribute_proxy, poll_period_sec):
    poll_period_msec = round(poll_period_sec * 1000)
    assert not attribute_proxy.is_polled()
    attribute_proxy.poll(poll_period_msec)
    assert attribute_proxy.is_polled()
    assert attribute_proxy.get_poll_period() == poll_period_msec


def _assert_polling_can_be_stopped(attribute_proxy):
    attribute_proxy.stop_poll()
    assert not attribute_proxy.is_polled()


def _write_values_and_read_via_polling(attribute_proxy, poll_period_sec, values):
    for value in values:
        attribute_proxy.write(value)
        time.sleep(poll_period_sec)
    assert len(values) <= DefaultPollRingDepth
    history = attribute_proxy.history(DefaultPollRingDepth)
    tolerance_for_slow_ci_runners = 2
    assert len(history) >= len(values) - tolerance_for_slow_ci_runners
    return history


def _assert_reading_times_increase_monotonically(history, t_start):
    t_previous = t_start
    for reading in history:
        t_current = reading.time.totime()
        assert t_current > t_previous
        t_previous = t_current


def _assert_reading_values_valid(history, initial_value, written_values):
    valid_values = _get_comparable_values([initial_value] + list(written_values))
    history_values = _get_comparable_values([reading.value for reading in history])
    last_index = -1
    for history_value in history_values:
        assert history_value in valid_values
        # check that historical values only move forward through the written values
        # i.e. polling buffer may repeat values, but may not return to an earlier value
        index = valid_values.index(history_value)
        assert index >= last_index
        last_index = index


def _get_comparable_values(values):
    comparable_values = []
    for value in values:
        if isinstance(value, tuple):
            value = list(value)
        elif isinstance(value, np.ndarray):
            value = value.tolist()
        comparable_values.append(value)
    return comparable_values


max_reply_attempts = 10
delay = 0.1


def test_read_write_attribute_async(attribute_proxy):
    value = TEST_VALUES[attribute_proxy.name()][0]
    w_id = attribute_proxy.write_asynch(value, wait=True)
    got_reply, attempt = False, 0
    while not got_reply:
        try:
            attribute_proxy.write_reply(w_id, wait=True)
            got_reply = True
        except DevFailed:
            attempt += 1
            if attempt >= max_reply_attempts:
                raise RuntimeError(
                    f"Test failed: cannot get write reply within {max_reply_attempts*delay} sec"
                )
            time.sleep(delay)

    r_id = attribute_proxy.read_asynch(wait=True)
    got_reply, attempt = False, 0
    while not got_reply:
        try:
            ret = attribute_proxy.read_reply(r_id, wait=True)
            got_reply = True
        except DevFailed:
            attempt += 1
            if attempt >= max_reply_attempts:
                raise RuntimeError(
                    f"Test failed: cannot get read reply within {max_reply_attempts*delay} sec"
                )
            time.sleep(delay)

    assert_close(ret.value, value)


class EasyEventDevice(Device):
    def init_device(self):
        self.set_change_event("attr", implemented=True, detect=False)

    @attribute
    def attr(self) -> int:
        return 1

    @command
    def send_event(self):
        self.push_change_event("attr", 2)


@pytest.mark.parametrize("green_mode", GreenMode.values.values(), ids=str)
def test_event(green_mode):
    with DeviceTestContext(EasyEventDevice, device_name="test/device/1", process=True):
        attr_proxy = attribute_proxy_map[green_mode]("test/device/1/attr")
        dev_proxy = attr_proxy.get_device_proxy()
        cb = (
            AsyncEventCallback() if green_mode == GreenMode.Asyncio else EventCallback()
        )
        eid = attr_proxy.subscribe_event(EventType.CHANGE_EVENT, cb, wait=True)
        dev_proxy.command_inout("send_event", wait=True)
        evts = cb.get_events()
        rep = 0
        while len(evts) < 2 and rep < 50:
            if green_mode in {GreenMode.Asyncio, GreenMode.Gevent}:
                # For asyncio and gevent green mode, the event callback is
                # scheduled on the event loop when it arrives.  We have to exercise
                # the event loop so that it gets a chance to invoke the callback.
                # One way to do that is sending a command that we wait for.
                # (For synchronous green mode, the callback will be invoked from
                # another thread so we can just sleep)
                # In a typical asyncio app, we would be awaiting other tasks, so
                # this wouldn't be necessary.
                dev_proxy.command_inout("state", wait=True)
            rep += 1
            evts = cb.get_events()
            time.sleep(0.1)
        if len(evts) < 2:
            pytest.fail(f"Cannot receive events in {green_mode}")
        assert_close([evt.attr_value.value for evt in evts[:2]], [1, 2])
        attr_proxy.unsubscribe_event(eid, wait=True)


@pytest.fixture
def uninitialized_attr_proxy():
    """
    This could happen if there was an exception in the AttributeProxy __init__ method,
    and the user is running through pytest.  pytest will have a reference to the frame
    and try to print out all the objects from the failed test.
    """
    proxy_instance = None
    try:
        AttributeProxy("not/existing/device/attr")
    except DevFailed:
        traceback = sys.exc_info()[2]
        for v in traceback.tb_next.tb_frame.f_locals.values():
            if isinstance(v, AttributeProxy):
                proxy_instance = v

    assert proxy_instance is not None
    yield proxy_instance


def test_pytest_report_on_failed_attribute_proxy_does_not_crash(
    uninitialized_attr_proxy,
):
    safe_repr = repr(uninitialized_attr_proxy)
    assert "AttributeProxy" in safe_repr
    assert "Unknown" in safe_repr

    safe_str = str(uninitialized_attr_proxy)
    assert "AttributeProxy" in safe_str
    assert "Unknown" in safe_str



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

try:
    from warnings import deprecated
except ImportError:
    from typing_extensions import deprecated

__all__ = ("loads", "dumps")


@deprecated(
    "loads function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def loads(fmt, data):
    if fmt.startswith("pickle"):
        import pickle

        loads = pickle.loads
    elif fmt.startswith("json"):
        import json

        loads = json.loads
    else:
        raise TypeError(f"Format '{fmt}' not supported")
    return loads(data)


@deprecated(
    "dumps function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def dumps(fmt, obj):
    if fmt.startswith("pickle"):
        import pickle

        ret = fmt, pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
        return ret
    elif fmt.startswith("json"):
        import json

        return fmt, json.dumps(obj)
    raise TypeError(f"Format '{fmt}' not supported")



import logging

import tango

from tango.server import Device
from tango.server import attribute, command


class IfchangeServer(Device):
    def init_device(self):
        Device.init_device(self)
        logging.basicConfig(level=logging.DEBUG)
        self._current = 0.0
        self.set_state(tango.DevState.ON)

    @attribute(label="Sequence Counter", dtype="int", description="Sequence counter")
    def seq_counter(self):
        return 456

    @attribute(label="Voltage", dtype="float", description="voltage")
    def volts(self):
        return 3.142

    def read_current(self):
        print("read current method ", self._current)
        return self._current

    def write_current(self, curr):
        self._current = curr
        print("set current to ", self._current)

    def start(self, argin):
        print("start method")
        return 3142

    @command(dtype_in=str, doc_in="name of dynamic attribute to add")
    def add_dyn_attr(self, name):
        attr = attribute(
            name=name, dtype="float", fget=self.read_current, fset=self.write_current
        )
        self.add_attribute(attr)

    @command(dtype_in=str, doc_in="name of dynamic attribute to delete")
    def delete_dyn_attr(self, name):
        self._remove_attribute(name)

    @command
    def add_dyn_cmd(self):
        device_level = True
        cmd = command(
            f=self.start,
            dtype_in=str,
            dtype_out=int,
            doc_in="description of input",
            doc_out="description of output",
            display_level=tango.DispLevel.EXPERT,
            polling_period=5100,
        )
        self.add_command(cmd, device_level)

    @command(dtype_in=str, doc_in="name of dynamic command to delete")
    def delete_dyn_cmd(self, name):
        self._remove_command(name, False, True)


# ----------
# Run server
# ----------
if __name__ == "__main__":
    IfchangeServer.run_server()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""Demo Tango Device Server using asyncio green mode"""

import asyncio
from tango import DevState, GreenMode
from tango.server import Device, command, attribute


class AsyncioDevice(Device):
    green_mode = GreenMode.Asyncio

    async def init_device(self):
        await super().init_device()
        self.set_state(DevState.ON)

    @command
    async def long_running_command(self):
        self.set_state(DevState.OPEN)
        await asyncio.sleep(2)
        self.set_state(DevState.CLOSE)

    @command
    async def background_task_command(self):
        loop = asyncio.get_event_loop()
        future = loop.create_task(self.coroutine_target())

    async def coroutine_target(self):
        self.set_state(DevState.INSERT)
        await asyncio.sleep(15)
        self.set_state(DevState.EXTRACT)

    @attribute
    async def test_attribute(self):
        await asyncio.sleep(2)
        return 42


if __name__ == "__main__":
    AsyncioDevice.run_server()



from tango import Util, DeviceProxy
from tango.server import Device, command
from tango.test_context import DeviceTestContext, MultiDeviceTestContext


class DeviceToTest1(Device):

    @command
    def check_device(self, device_instance: str):
        assert (
            Util.instance().get_sub_dev_diag().get_associated_device()
            == device_instance
        )


devices_info = (
    {
        "class": DeviceToTest1,
        "devices": [{"name": "test/device/1"}, {"name": "test/device/2"}],
    },
)


def test_associated_device_is_device_instance():

    # NOTE: this test works only with process=True
    with MultiDeviceTestContext(devices_info, process=True):
        dev1 = DeviceProxy("test/device/1")
        dev2 = DeviceProxy("test/device/2")

        dev1.check_device("test/device/1")
        dev2.check_device("test/device/2")


class DeviceToTest2(Device):

    sdd = None

    def init_device(self):
        super().init_device()
        self.sdd = Util.instance().get_sub_dev_diag()

    @command
    def add_device(self, device_instance: str):
        self.sdd.register_sub_device(device_instance, "test_device")

    @command
    def remove_device(self, device_instance: str):
        self.sdd.remove_sub_devices(device_instance)

    @command
    def check_device(self, test_string: str):
        if len(test_string) > 0:
            assert self.sdd.get_sub_devices() == [test_string]
        else:
            assert self.sdd.get_sub_devices() == []


def test_register_sub_device():

    with DeviceTestContext(
        DeviceToTest2, device_name="test/device/1", process=True
    ) as dev:
        dev.check_device("")
        dev.add_device("test/device/1")
        dev.check_device("test/device/1 test_device")
        dev.remove_device("test/device/1")
        dev.check_device("")



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# client example: writing an array
import sys
from tango import DeviceProxy, DevFailed, Except

dev_name = "sys/tg_test/1"
attr_name = "double_spectrum"

# step 1 create the device proxy
try:
    dev = DeviceProxy(dev_name)
    print(f"proxy for {dev_name} created")
except DevFailed as ex:
    Except.print_exception(ex)
    sys.exit(1)

# prepare value list
values = []
for i in range(0, 255, 1):
    values.append(i / 10.0)
try:
    dev.write_attribute(attr_name, values)
    print(f"write_attribute {attr_name!r} OK")
except DevFailed as ex:
    Except.print_exception(ex)
    sys.exit(1)

# use numpy.ndarray - must import numpy
import numpy

arr_values = numpy.empty(255)
for i in range(0, 255, 1):
    arr_values[i] = i / 20.0
try:
    dev.write_attribute(attr_name, arr_values)
    dev.float_spectrum = arr_values  # simpler syntax on another attribute
    print(f"write_attribute {attr_name!r} with numpy.ndarray OK")
except DevFailed as ex:
    Except.print_exception(ex)
    sys.exit(1)

"""
Typical output:

➜  training git:(develop) ✗ docker-compose exec cli /training/client/writing02.py
proxy for sys/tg_test/1 created
write_attribute 'double_spectrum' OK
write_attribute 'double_spectrum' with numpy.ndarray OK
"""



import time
from datetime import datetime

from tango import TimeVal
from tango.test_utils import assert_close


def test_timeval():
    # test constructors
    from_default = TimeVal()
    assert from_default.tv_sec == 0
    assert from_default.tv_usec == 0
    assert from_default.tv_nsec == 0

    from_ints = TimeVal(1, 2, 3)
    assert from_ints.tv_sec == 1
    assert from_ints.tv_usec == 2
    assert from_ints.tv_nsec == 3

    from_int = TimeVal(100)
    assert from_int.tv_sec == 100
    assert from_int.tv_usec == 0
    assert from_int.tv_nsec == 0

    from_float = TimeVal(100.200300400)
    assert from_float.tv_sec == 100
    assert from_float.tv_usec == 200300
    assert from_float.tv_nsec == 400

    from_datetime = TimeVal(datetime.now())
    assert from_datetime.tv_sec > 0

    # test convertors:
    now_float = time.time()
    now_tv = TimeVal(now_float)
    assert_close(now_float, now_tv.totime())

    now_datetime = datetime.now()
    now_tv = TimeVal(now_datetime)
    assert now_tv.todatetime() == now_datetime
